from typing import List, Dict, Iterable, Optional

import torch
import torch.nn.functional as F

from lightwood.encoder import BaseEncoder
from lightwood.encoder.categorical import OneHotEncoder


class TsCatArrayEncoder(BaseEncoder):
    def __init__(self, timesteps: int, is_target: bool = False, grouped_by=None):
        """
        This encoder handles arrays of categorical time series data by wrapping the OHE encoder with behavior specific to time series tasks.

        :param timesteps: length of forecasting horizon, as defined by TimeseriesSettings.window.
        :param is_target: whether this encoder corresponds to the target column.
        :param grouped_by: what columns, if any, are considered to group the original column and yield multiple time series.
        """  # noqa
        super(TsCatArrayEncoder, self).__init__(is_target=is_target)
        self.group_combinations = None
        self.dependencies = grouped_by
        self.data_window = timesteps
        self.sub_encoder = OneHotEncoder(is_target=is_target, use_unknown=False)

    def prepare(self, priming_data):
        """
        This method prepares the underlying time series numerical encoder.
        """
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self.sub_encoder.prepare(priming_data)
        self.output_size = self.data_window * self.sub_encoder.output_size
        self.rev_map = self.sub_encoder.rev_map
        self.is_prepared = True

    def encode(self, data: Iterable[Iterable], dependency_data: Optional[Dict[str, str]] = {}) -> torch.Tensor:
        """
        Encodes a list of time series arrays using the underlying time series numerical encoder.

        :param data: list of numerical values to encode. Its length is determined by the tss.window parameter, and all data points belong to the same time series.
        :param dependency_data: dict with values of each group_by column for the time series, used to retrieve the correct normalizer.

        :return: list of encoded time series arrays. Tensor is (len(data), N x K)-shaped, where N: self.data_window and K: sub-encoder # of output features.
        """  # noqa
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        ret = []
        for series in data:
            ret.append(self.encode_one(series))

        return torch.vstack(ret)

    def encode_one(self, data: Iterable) -> torch.Tensor:
        """
        Encodes a single windowed slice of any given time series.

        :param data: windowed slice of a numerical time series.

        :return: an encoded time series array, as per the underlying `TsNumericEncoder` object. 
        The output of this encoder for all time steps is concatenated, so the final shape of the tensor is (1, NxK), where N: self.data_window and K: sub-encoder # of output features. 
        """  # noqa
        ret = []

        for data_point in data:
            ret.append(self.sub_encoder.encode([data_point]))

        ret = torch.hstack(ret)
        padding_size = self.output_size - ret.shape[-1]

        if padding_size > 0:
            ret = F.pad(ret, (0, padding_size))

        return ret

    def decode(self, encoded_values, dependency_data=None) -> List[List]:
        """
        Decodes a list of encoded arrays into values in their original domains.

        :param encoded_values: encoded slices of numerical time series.
        :param dependency_data: used to determine the correct normalizer for the input.

        :return: a list of decoded time series arrays.
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        encoded_values = encoded_values.reshape(encoded_values.shape[0],
                                                self.data_window,
                                                self.sub_encoder.output_size)

        ret = []
        for tensor in torch.split(encoded_values, 1, dim=0):
            ret.append(self.decode_one(tensor))

        return ret

    def decode_one(self, encoded_value) -> List:
        """
        Decodes a single window of a time series into its original domain.

        :param encoded_value: encoded slice of a numerical time series.
        :param dependency_data: used to determine the correct normalizer for the input.

        :return: a list of length TimeseriesSettings.window with decoded values for the forecasted time series.
        """
        ret = []
        for encoded_timestep in torch.split(encoded_value, 1, dim=1):
            ret.extend(self.sub_encoder.decode(encoded_timestep.squeeze(1)))
        return ret



from typing import Dict, List, Optional, Union
import time

import torch
import optuna
import xgboost as xgb
import numpy as np
import pandas as pd
from sklearn.preprocessing import OrdinalEncoder
from type_infer.dtype import dtype

from lightwood.helpers.log import log
from lightwood.helpers.parallelism import get_nr_procs
from lightwood.mixer.base import BaseMixer
from lightwood.encoder.base import BaseEncoder
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs

optuna.logging.set_verbosity(optuna.logging.CRITICAL)


def check_gpu_support():
    try:
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split
        data = load_iris()
        x, _, y, _ = train_test_split(data['data'], data['target'], test_size=.2)
        bst = xgb.XGBClassifier(n_estimators=2, max_depth=2, learning_rate=1, objective='binary:logistic',
                                tree_method='gpu_hist', gpu_id=0)
        bst.fit(x, y)
        return True
    except xgb.core.XGBoostError:
        return False


class XGBoostMixer(BaseMixer):
    model: Union[xgb.XGBClassifier, xgb.XGBRegressor]
    ordinal_encoder: OrdinalEncoder
    label_set: List[str]
    max_bin: int
    device: torch.device
    device_str: str
    num_iterations: int
    use_optuna: bool
    supports_proba: bool

    """
    Gradient boosting mixer with an XGBoost backbone.

    This mixer is a good all-rounder, due to the generally great performance of tree-based ML algorithms for supervised learning tasks with tabular data.
    If you want more information regarding the techniques that set apart XGBoost from other gradient boosters, please refer to their technical paper: "XGBoost: A Scalable Tree Boosting System" (2016).

    We can basically think of this mixer as a wrapper to the XGBoost Python package. To do so, there are a few caveats the user may want to be aware about:
        * If you seek GPU utilization, XGBoost must be compiled from source instead of being installed through `pip`.
        * Integer, float, and quantity `dtype`s are treated as regression tasks with `reg:squarederror` loss. All other supported `dtype`s is casted as a multiclass task with `multi:softmax` loss.
        * A partial fit can be performed with the `dev` data split as part of `fit`, if specified with the `fit_on_dev` argument.

    There are a couple things in the backlog that will hopefully be added soon:
        * An automatic optuna-based hyperparameter search. This procedure triggers when a single iteration of XGBoost is deemed fast enough (given the time budget).
        * Support for "unknown class" as a possible answer for multiclass tasks.
    """  # noqa

    def __init__(
            self, stop_after: float, target: str, dtype_dict: Dict[str, str],
            input_cols: List[str],
            fit_on_dev: bool, use_optuna: bool, target_encoder: BaseEncoder):
        """
        :param stop_after: time budget in seconds.
        :param target: name of the target column that the mixer will learn to predict.
        :param dtype_dict: dictionary with dtypes of all columns in the data.
        :param input_cols: list of column names.
        :param fit_on_dev: whether to perform a `partial_fit()` at the end of `fit()` using the `dev` data split.
        :param use_optuna: whether to activate the automated hyperparameter search (optuna-based). Note that setting this flag to `True` does not guarantee the search will run, rather, the speed criteria will be checked first (i.e., if a single iteration is too slow with respect to the time budget, the search will not take place). 
        """  # noqa
        super().__init__(stop_after)
        self.model = None
        self.ordinal_encoder = None
        self.positive_domain = False
        self.label_set = []
        self.target = target
        self.dtype_dict = dtype_dict
        self.input_cols = input_cols
        self.use_optuna = use_optuna
        self.params = {}
        self.fit_on_dev = fit_on_dev
        self.cls_dtypes = [dtype.categorical, dtype.binary, dtype.cat_tsarray]
        self.float_dtypes = [dtype.float, dtype.quantity, dtype.num_tsarray]
        self.num_dtypes = [dtype.integer] + self.float_dtypes
        self.supports_proba = dtype_dict[target] in self.cls_dtypes
        self.stable = True
        self.target_encoder = target_encoder

        gpu_works = check_gpu_support()
        if not gpu_works:
            self.device = torch.device('cpu')
            self.device_str = 'cpu'
            log.warning('XGBoost running on CPU')
        else:
            self.device = torch.device('cuda')
            self.device_str = 'gpu'

        self.max_bin = 255

    def _to_dataset(self, ds: EncodedDs, output_dtype: str, mode='train'):
        """
        Helper method to wrangle a datasource into the format that the underlying model requires.
        :param ds: EncodedDS that contains the dataframe to transform.
        :param output_dtype:
        :return: modified `data` object that conforms to XGBoost's expected format.
        """  # noqa
        data = None
        for input_col in self.input_cols:
            if data is None:
                data = ds.get_encoded_column_data(input_col).to(self.device)
            else:
                enc_col = ds.get_encoded_column_data(input_col).to(self.device)
                data = torch.cat((data, enc_col.to(self.device)), 1)

        data = data.cpu().numpy()

        if mode in ('train', 'dev'):
            weights = []
            label_data = ds.get_column_original_data(self.target)
            if output_dtype in self.cls_dtypes:
                if mode == 'train':
                    self.ordinal_encoder = OrdinalEncoder()
                    self.label_set = list(set(label_data))
                    self.ordinal_encoder.fit(np.array(list(self.label_set)).reshape(-1, 1))

                filtered_label_data = []
                for x in label_data:
                    if x in self.label_set:
                        filtered_label_data.append(x)

                weight_map = getattr(self.target_encoder, 'target_weights', None)
                if weight_map is not None:
                    weights = [weight_map[x] for x in label_data]

                label_data = self.ordinal_encoder.transform(np.array(filtered_label_data).reshape(-1, 1)).flatten()

            elif output_dtype in self.num_dtypes:
                weight_map = getattr(self.target_encoder, 'target_weights', None)
                if weight_map is not None:
                    target_encoder = ds.encoders[self.target]

                    # get the weights from the numeric target encoder
                    weights = target_encoder.get_weights(label_data)

                if output_dtype in self.float_dtypes:
                    label_data = label_data.astype(float)
                elif output_dtype == dtype.integer:
                    label_data = label_data.clip(-pow(2, 63), pow(2, 63)).astype(int)

            return data, label_data, weights

        else:
            return data

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the XGBoost model.

        :param train_data: encoded features for training dataset
        :param dev_data: encoded features for dev dataset
        """
        started = time.time()

        log.info('Started fitting XGBoost model')
        self.fit_data_len = len(train_data)
        self.positive_domain = getattr(train_data.encoders.get(self.target, None), 'positive_domain', False)
        output_dtype = self.dtype_dict[self.target]

        if output_dtype not in self.cls_dtypes + self.num_dtypes:
            log.error(f'XGBoost mixer not supported for type: {output_dtype}')
            raise Exception(f'XGBoost mixer not supported for type: {output_dtype}')
        else:
            objective = 'reg:squarederror' if output_dtype in self.num_dtypes else 'multi:softprob'
            metric = 'rmse' if output_dtype in self.num_dtypes else 'mlogloss'

        self.params = {
            'objective': objective,
            'eval_metric': metric,
            'n_jobs': get_nr_procs(train_data.data_frame),
            'process_type': 'default',  # normal training
            'verbosity': 0,
            'early_stopping_rounds': 5
            # 'device_type': self.device_str,  # TODO
        }

        # Prepare the data
        train_dataset, train_labels, train_weights = self._to_dataset(train_data, output_dtype, mode='train')
        dev_dataset, dev_labels, dev_weights = self._to_dataset(dev_data, output_dtype, mode='dev')

        if output_dtype not in self.num_dtypes:
            self.all_classes = self.ordinal_encoder.categories_[0]
            self.params['num_class'] = self.all_classes.size
            model_class = xgb.XGBClassifier
        else:
            model_class = xgb.XGBRegressor

        # Determine time per iterations
        start = time.time()
        self.params['n_estimators'] = 1

        with xgb.config_context(verbosity=0):
            self.model = model_class(**self.params)
            if train_weights is not None and dev_weights is not None:
                self.model.fit(train_dataset, train_labels, sample_weight=train_weights,
                               eval_set=[(dev_dataset, dev_labels)],
                               sample_weight_eval_set=[dev_weights])
            else:
                self.model.fit(train_dataset, train_labels,
                               eval_set=[(dev_dataset, dev_labels)])

        end = time.time()
        seconds_for_one_iteration = max(0.1, end - start)

        self.stop_after = max(1, self.stop_after - (time.time() - started))

        # Determine nr of iterations
        log.info(f'A single GBM iteration takes {seconds_for_one_iteration} seconds')
        self.num_iterations = int(self.stop_after * 0.8 / seconds_for_one_iteration)

        # TODO: Turn on grid search if training doesn't take too long using it
        # kwargs = {}
        # if self.use_optuna and self.num_iterations >= 200:
        #     model_generator = None
        #     kwargs['time_budget'] = self.stop_after * 0.4
        #     self.num_iterations = int(self.num_iterations / 2)
        #     kwargs['optuna_seed'] = 0
        # else:
        #     model_generator = None

        # Train the models
        log.info(
            f'Training XGBoost with {self.num_iterations} iterations given {self.stop_after} seconds constraint')  # noqa
        if self.num_iterations < 1:
            self.num_iterations = 1
        self.params['n_estimators'] = int(self.num_iterations)

        # TODO: reinstance these on each run! keep or use via eval_set in fit()?
        # self.params['callbacks'] = [(train_dataset, 'train'), (dev_dataset, 'eval')]

        with xgb.config_context(verbosity=0):
            self.model = model_class(**self.params)
            if train_weights is not None and dev_weights is not None:
                self.model.fit(train_dataset, train_labels, sample_weight=train_weights,
                               eval_set=[(dev_dataset, dev_labels)],
                               sample_weight_eval_set=[dev_weights])
            else:
                self.model.fit(train_dataset, train_labels,
                               eval_set=[(dev_dataset, dev_labels)])

        if self.fit_on_dev:
            self.partial_fit(dev_data, train_data)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        # TODO: https://xgboost.readthedocs.io/en/stable/python/python_api.html#module-xgboost.sklearn
        # To resume training from a previous checkpoint, explicitly pass xgb_model argument.
        log.info('XGBoost mixer does not have a `partial_fit` implementation')

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Call a trained XGBoost mixer to output predictions for the target column.

        :param ds: input data with values for all non-target columns.
        :param args: inference-time arguments (e.g. whether to output predicted labels or probabilities).

        :return: dataframe with predictions.
        """  # noqa
        output_dtype = self.dtype_dict[self.target]
        xgbdata = self._to_dataset(ds, self.dtype_dict[self.target], mode='predict')

        with xgb.config_context(verbosity=0):
            if output_dtype in self.num_dtypes:
                raw_predictions = self.model.predict(xgbdata)
            else:
                raw_predictions = self.model.predict_proba(xgbdata)

        if self.ordinal_encoder is not None:
            decoded_predictions = self.ordinal_encoder.inverse_transform(
                np.argmax(raw_predictions, axis=1).reshape(-1, 1)).flatten()
        else:
            decoded_predictions = raw_predictions

        if self.positive_domain:
            decoded_predictions = [max(0, p) for p in decoded_predictions]

        ydf = pd.DataFrame({'prediction': decoded_predictions})

        if args.predict_proba and self.ordinal_encoder is not None:
            for idx, label in enumerate(self.ordinal_encoder.categories_[0].tolist()):
                ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

        return ydf



from copy import deepcopy
from typing import Dict, Tuple
from types import SimpleNamespace

import torch
import pandas as pd
from torch import nn, optim
from sklearn.preprocessing import OrdinalEncoder

from lightwood.helpers.log import log
from lightwood.analysis.base import BaseAnalysisBlock


class TempScaler(BaseAnalysisBlock):
    """
    Original reference (MIT Licensed): https://github.com/gpleiss/temperature_scaling
    NB: Output of the neural network should be the classification logits, NOT the softmax (or log softmax)! TODO
    """
    def __init__(self, deps=tuple()):
        super().__init__(deps=deps)
        self.temperature = nn.Parameter(torch.ones(1))
        self.ordenc = OrdinalEncoder()
        self._softmax = torch.nn.Softmax(dim=1)
        self.active = False

    def temperature_scale(self, logits):
        temperature = self.temperature.unsqueeze(1).expand(logits.size(0), logits.size(1))
        return logits / temperature

    def softmax(self, logits):
        return self._softmax(self.temperature_scale(logits))

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        """
        Tune and set the temperature of a neural model optimizing NLL using validation set logits.
        """
        ns = SimpleNamespace(**kwargs)

        if ns.predictor.mixers[ns.predictor.indexes_by_accuracy[0]].supports_proba:
            self.n_cls = len(ns.stats_info.train_observed_classes)
            nll_criterion = nn.CrossEntropyLoss()
            self.ordenc.fit([[val] for val in ns.stats_info.train_observed_classes])

            # collect logits and labels for the validation set
            logits_list = []
            labels_list = []
            with torch.no_grad():
                prob_cols = [col for col in ns.normal_predictions.columns
                             if '__mdb_proba' in col and
                             '__mdb_unknown_cat' not in col]
                if not prob_cols:
                    return info  # early stop if no proba info is available
                for logits, label in zip(ns.normal_predictions[prob_cols].values, ns.data[ns.target]):
                    logits_list.append(logits.tolist())
                    labels_list.append(int(self.ordenc.transform([[label]]).flatten()[0]))
                logits = torch.tensor(logits_list)
                labels = torch.tensor(labels_list).long()

            # NLL and ECE before temp scaling
            before_temperature_nll = nll_criterion(logits, labels).item()
            log.info(f'Before calibration - NLL: {round(before_temperature_nll, 3)}')

            # optimize w.r.t. NLL
            optimizer = optim.LBFGS([self.temperature], lr=0.001, max_iter=1000)

            def eval_loss():
                optimizer.zero_grad()
                loss = nll_criterion(self.temperature_scale(logits), labels)
                loss.backward()
                return loss

            optimizer.step(eval_loss)

            # NLL and ECE after temp scaling
            after_temperature_nll = nll_criterion(self.temperature_scale(logits), labels).item()
            log.info(f'Optimal temperature: {round(self.temperature.item(), 3)}')
            log.info(f'After calibration - NLL: {round(after_temperature_nll, 3)}')

            output = deepcopy(ns.normal_predictions)
            output['confidence'] = torch.max(self.softmax(logits), dim=1).values.detach().numpy()
            info['result_df'] = output
        self.active = True
        return info

    def explain(self,
                row_insights: pd.DataFrame,
                global_insights: Dict[str, object], **kwargs) -> Tuple[pd.DataFrame, Dict[str, object]]:
        """ Perform temperature scaling on logits """
        prob_cols = [col for col in row_insights.columns
                     if '__mdb_proba' in col and
                     '__mdb_unknown_cat' not in col]
        if self.active and prob_cols:
            logits = torch.tensor(row_insights[prob_cols].values)
            confs = self.softmax(logits)
            row_insights['confidence'] = torch.max(confs, dim=1).values.detach().numpy().reshape(-1, 1)
        else:
            row_insights['confidence'] = torch.max(
                torch.tensor(row_insights[prob_cols].values), dim=1).values.detach().numpy().reshape(-1, 1)
        return row_insights, global_insights




"""
Nonconformity functions.
"""

# Original author: Henrik Linusson (github.com/donlnz)

import abc
import numpy as np
import sklearn.base
from scipy.interpolate import interp1d
from copy import deepcopy


# -----------------------------------------------------------------------------
# Error functions
# -----------------------------------------------------------------------------
class ClassificationErrFunc(object):
    """Base class for classification model error functions.
    """ # noqa

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(ClassificationErrFunc, self).__init__()

    @abc.abstractmethod
    def apply(self, prediction, y):
        """Apply the nonconformity function.

        Parameters
        ----------
        prediction : numpy array of shape [n_samples, n_classes]
            Class probability estimates for each sample.

        y : numpy array of shape [n_samples]
            True output labels of each sample.

        Returns
        -------
        nc : numpy array of shape [n_samples]
            Nonconformity scores of the samples.
        """ # noqa
        pass


class RegressionErrFunc(object):
    """Base class for regression model error functions.
    """ # noqa

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(RegressionErrFunc, self).__init__()

    @abc.abstractmethod
    def apply(self, prediction, y):  # , norm=None, beta=0):
        """Apply the nonconformity function.

        Parameters
        ----------
        prediction : numpy array of shape [n_samples, n_classes]
            Class probability estimates for each sample.

        y : numpy array of shape [n_samples]
            True output labels of each sample.

        Returns
        -------
        nc : numpy array of shape [n_samples]
            Nonconformity scores of the samples.
        """ # noqa
        pass

    @abc.abstractmethod
    def apply_inverse(self, nc, significance):  # , norm=None, beta=0):
        """Apply the inverse of the nonconformity function (i.e.,
        calculate prediction interval).

        Parameters
        ----------
        nc : numpy array of shape [n_calibration_samples]
            Nonconformity scores obtained for conformal predictor.

        significance : float
            Significance level (0, 1).

        Returns
        -------
        interval : numpy array of shape [n_samples, 2]
            Minimum and maximum interval boundaries for each prediction.
        """ # noqa
        pass


class TSErrFunc(object):
    """Base class for time series model error functions.
    """ # noqa

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(TSErrFunc, self).__init__()

    @abc.abstractmethod
    def apply(self, prediction, y):
        """Apply the nonconformity function.

        Parameters
        ----------
        prediction : numpy array of shape [n_samples, horizon_length]
            Forecasts for each sample.

        y : numpy array of shape [n_samples, horizon_length]
            True output series for each sample.

        Returns
        -------
        nc : numpy array of shape [n_samples, horizon_length]
            Nonconformity scores of the samples.
        """ # noqa
        pass

    @abc.abstractmethod
    def apply_inverse(self, nc, significance):  # , norm=None, beta=0):
        """Apply the inverse of the nonconformity function (i.e.,
        calculate prediction interval).

        Parameters
        ----------
        nc : numpy array of shape [n_calibration_samples, horizon_length]
            Nonconformity scores obtained for conformal predictor.

        significance : float
            Significance level (0, 1).

        Returns
        -------
        interval : numpy array of shape [n_samples, horizon_length, 2]
            Minimum and maximum interval boundaries for each step of the forecast.
        """ # noqa
        pass


class InverseProbabilityErrFunc(ClassificationErrFunc):
    """Calculates the probability of not predicting the correct class.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        1 - hat{P}(y_i | x) , .
    """ # noqa

    def __init__(self):
        super(InverseProbabilityErrFunc, self).__init__()

    def apply(self, prediction, y):
        prob = np.zeros(y.size, dtype=np.float32)
        for i, y_ in enumerate(y):
            if y_ >= prediction.shape[1]:
                prob[i] = 0
            else:
                prob[i] = prediction[i, int(y_)]
        return 1 - prob


class MarginErrFunc(ClassificationErrFunc):
    """
    Calculates the margin error.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        0.5 - frac{hat{P}(y_i | x) - max_{y , != , y_i} hat{P}(y | x)}{2}
    """ # noqa

    def __init__(self):
        super(MarginErrFunc, self).__init__()

    def apply(self, prediction, y):
        prediction = deepcopy(prediction).astype(float)
        prob = np.zeros(y.size, dtype=np.float32)
        for i, y_ in enumerate(y):
            if y_ >= prediction.shape[1]:
                prob[i] = 0
            else:
                prob[i] = prediction[i, int(y_)]
                prediction[i, int(y_)] = -np.inf
        return 0.5 - ((prob - prediction.max(axis=1)) / 2)


class AbsErrorErrFunc(RegressionErrFunc):
    """Calculates absolute error nonconformity for regression problems.

        For each correct output in ``y``, nonconformity is defined as

        .. math::
            | y_i - hat{y}_i |
    """ # noqa

    def __init__(self):
        super(AbsErrorErrFunc, self).__init__()

    def apply(self, prediction, y):
        return np.abs(prediction - y)

    def apply_inverse(self, nc, significance):
        nc = np.sort(nc)[::-1]
        border = int(np.floor(significance * (nc.size + 1))) - 1
        # TODO: should probably warn against too few calibration examples
        border = min(max(border, 0), nc.size - 1)
        return np.vstack([nc[border], nc[border]])


class BoostedAbsErrorErrFunc(RegressionErrFunc):
    """ Calculates absolute error nonconformity for regression problems. Applies linear interpolation
    for nonconformity scores when we have less than 100 samples in the validation dataset.
    """ # noqa

    def __init__(self):
        super(BoostedAbsErrorErrFunc, self).__init__()

    def apply(self, prediction, y):
        return np.abs(prediction - y)

    def apply_inverse(self, nc, significance):
        nc = np.sort(nc)[::-1]
        border = int(np.floor(significance * (nc.size + 1))) - 1
        if 1 < nc.size < 100:
            x = np.arange(nc.shape[0])
            interp = interp1d(x, nc)
            nc = interp(np.linspace(0, nc.size - 1, 100))
        border = min(max(border, 0), nc.size - 1)
        return np.vstack([nc[border], nc[border]])


class SignErrorErrFunc(RegressionErrFunc):
    """Calculates signed error nonconformity for regression problems.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        y_i - hat{y}_i

    References
    ----------
    .. [1] Linusson, Henrik, Ulf Johansson, and Tuve Lofstrom.
        Signed-error conformal regression. Pacific-Asia Conference on Knowledge
        Discovery and Data Mining. Springer International Publishing, 2014.
    """ # noqa

    def __init__(self):
        super(SignErrorErrFunc, self).__init__()

    def apply(self, prediction, y):
        return (prediction - y)

    def apply_inverse(self, nc, significance):
        nc = np.sort(nc)[::-1]
        upper = int(np.floor((significance / 2) * (nc.size + 1)))
        lower = int(np.floor((1 - significance / 2) * (nc.size + 1)))
        # TODO: should probably warn against too few calibration examples
        upper = min(max(upper, 0), nc.size - 1)
        lower = max(min(lower, nc.size - 1), 0)
        return np.vstack([-nc[lower], nc[upper]])


class TSAbsErrorErrFunc(TSErrFunc):
    """Calculates absolute error nonconformity for time series problems.

        For each forecasted step ``y_h`` for h \in 1..horizon, nonconformity is defined as

        .. math::
            | y_ni - hat{y}_ni |
            
        Following Stankeviciute, K. (2021). Conformal Time-Series Forecasting, we perform a 
        Bonferroni correction over the nonoconformity scores when applying the inverse function.
    """ # noqa

    def __init__(self, horizon_length):
        super(TSAbsErrorErrFunc, self).__init__()
        self.horizon_length = horizon_length

    def apply(self, prediction, y):
        """ calculate absolute error, eq. (6) in the paper """  # noqa
        return np.abs(prediction - y)

    def apply_inverse(self, nc, significance):
        significance /= self.horizon_length  # perform Bonferroni correction, eq. (7) in the paper
        nc = np.sort(nc, axis=0)[::-1]
        border = int(np.floor(significance * (nc.shape[0] + 1))) - 1
        border = min(max(border, 0), nc.shape[0] - 1)
        return nc[border]


# -----------------------------------------------------------------------------
# Base nonconformity scorer
# -----------------------------------------------------------------------------
class BaseScorer(sklearn.base.BaseEstimator):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(BaseScorer, self).__init__()

    @abc.abstractmethod
    def fit(self, x, y):
        pass

    @abc.abstractmethod
    def score(self, x, y=None):
        pass


class RegressorNormalizer(BaseScorer):
    def __init__(self, base_model, normalizer_model, err_func):
        super(RegressorNormalizer, self).__init__()
        self.base_model = base_model
        self.normalizer_model = normalizer_model
        self.err_func = err_func

    def fit(self, x, y):
        residual_prediction = self.base_model.predict(x)
        residual_error = np.abs(self.err_func.apply(residual_prediction, y))
        residual_error += 0.00001  # Add small term to avoid log(0)
        log_err = np.log(residual_error)
        self.normalizer_model.fit(x, log_err)

    def score(self, x, y=None):
        norm = np.exp(self.normalizer_model.predict(x))
        return norm


class BaseModelNc(BaseScorer):
    """Base class for nonconformity scorers based on an underlying model.

    Parameters
    ----------
    model : ClassifierAdapter or RegressorAdapter or TSAdapter
        Underlying classification model used for calculating nonconformity
        scores.

    err_func : ClassificationErrFunc or RegressionErrFunc or TSErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.
    """ # noqa

    def __init__(self, model, err_func, normalizer=None, beta=0):
        super(BaseModelNc, self).__init__()
        self.err_func = err_func
        self.model = model
        self.normalizer = normalizer
        self.beta = beta

        # If we use sklearn.base.clone (e.g., during cross-validation),
        # object references get jumbled, so we need to make sure that the
        # normalizer has a reference to the proper model adapter, if applicable.
        if (self.normalizer is not None and
                hasattr(self.normalizer, 'base_model')):
            self.normalizer.base_model = self.model

        self.last_x, self.last_y = None, None
        self.last_prediction = None
        self.clean = False

    def fit(self, x, y):
        """Fits the underlying model of the nonconformity scorer.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for fitting the underlying model.

        y : numpy array of shape [n_samples]
            Outputs of examples for fitting the underlying model.

        Returns
        -------
        None
        """ # noqa
        self.model.fit(x, y)
        if self.normalizer is not None:
            self.normalizer.fit(x, y)
        self.clean = False

    def score(self, x, y=None):
        """Calculates the nonconformity score of a set of samples.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for which to calculate a nonconformity score.

        y : numpy array of shape [n_samples]
            Outputs of examples for which to calculate a nonconformity score.

        Returns
        -------
        nc : numpy array of shape [n_samples]
            Nonconformity scores of samples.
        """ # noqa
        prediction = self.model.predict(x)

        err = self.err_func.apply(prediction, y)
        if self.normalizer is not None:
            try:
                norm = self.normalizer.score(x) + self.beta
                err = err / norm
            except Exception:
                pass

        return err

    def __deepcopy__(self, memo={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ['model', 'normalizer']:  # model should not be copied
                setattr(result, k, deepcopy(v, memo))
            else:
                setattr(result, k, v)
        return result


# -----------------------------------------------------------------------------
# Classification nonconformity scorers
# -----------------------------------------------------------------------------
class ClassifierNc(BaseModelNc):
    """Nonconformity scorer using an underlying class probability estimating
    model.

    Parameters
    ----------
    model : ClassifierAdapter
        Underlying classification model used for calculating nonconformity
        scores.

    err_func : ClassificationErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.

    Attributes
    ----------
    model : ClassifierAdapter
        Underlying model object.

    err_func : ClassificationErrFunc
        Scorer function used to calculate nonconformity scores.
    """ # noqa

    def __init__(self,
                 model,
                 err_func=MarginErrFunc(),
                 normalizer=None,
                 beta=0):
        super(ClassifierNc, self).__init__(model,
                                           err_func,
                                           normalizer,
                                           beta)


# -----------------------------------------------------------------------------
# Regression nonconformity scorers
# -----------------------------------------------------------------------------
class RegressorNc(BaseModelNc):
    """Nonconformity scorer using an underlying regression model.

    Parameters
    ----------
    model : RegressorAdapter
        Underlying regression model used for calculating nonconformity scores.

    err_func : RegressionErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.

    Attributes
    ----------
    model : RegressorAdapter
        Underlying model object.

    err_func : RegressionErrFunc
        Scorer function used to calculate nonconformity scores.
    """ # noqa

    def __init__(self,
                 model,
                 err_func=AbsErrorErrFunc(),
                 normalizer=None,
                 beta=0):
        super(RegressorNc, self).__init__(model,
                                          err_func,
                                          normalizer,
                                          beta)

    def predict(self, x, nc, significance=None):
        """Constructs prediction intervals for a set of test examples.

        Predicts the output of each test pattern using the underlying model,
        and applies the (partial) inverse nonconformity function to each
        prediction, resulting in a prediction interval for each test pattern.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        significance : float
            Significance level (maximum allowed error rate) of predictions.
            Should be a float between 0 and 1. If ``None``, then intervals for
            all significance levels (0.01, 0.02, ..., 0.99) are output in a
            3d-matrix.

        Returns
        -------
        p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99]
            If significance is ``None``, then p contains the interval (minimum
            and maximum boundaries) for each test pattern, and each significance
            level (0.01, 0.02, ..., 0.99). If significance is a float between
            0 and 1, then p contains the prediction intervals (minimum and
            maximum	boundaries) for the set of test patterns at the chosen
            significance level.
        """ # noqa
        n_test = x.shape[0]
        prediction = self.model.predict(x)

        norm = 1
        if self.normalizer is not None:
            try:
                norm = self.normalizer.score(x) + self.beta
            except Exception:
                pass

        if significance:
            err_dist = self.err_func.apply_inverse(nc, significance)
            err_dist = np.hstack([err_dist] * n_test)
            err_dist *= norm

            intervals = np.zeros((x.shape[0], 2))
            intervals[:, 0] = prediction - err_dist[0, :]
            intervals[:, 1] = prediction + err_dist[1, :]

            return intervals
        else:
            significance = np.arange(0.01, 1.0, 0.01)
            intervals = np.zeros((x.shape[0], 2, significance.size))

            for i, s in enumerate(significance):
                err_dist = self.err_func.apply_inverse(nc, s)
                err_dist = np.hstack([err_dist] * n_test)
                err_dist *= norm

                intervals[:, 0, i] = prediction - err_dist[0, :]
                intervals[:, 1, i] = prediction + err_dist[0, :]

            return intervals


# -----------------------------------------------------------------------------
# Time series nonconformity scorers
# -----------------------------------------------------------------------------
class TSNc(BaseModelNc):
    """Nonconformity scorer using an underlying time series model.

    Parameters
    ----------
    model : TSAdapter
        Underlying regression model used for calculating nonconformity scores.

    err_func : TSErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.

    Attributes
    ----------
    model : TSAdapter
        Underlying model object.

    err_func : TSErrFunc
        Scorer function used to calculate nonconformity scores.
    """ # noqa

    def __init__(self,
                 model,
                 err_func=TSAbsErrorErrFunc(horizon_length=1),
                 normalizer=None,
                 beta=0):
        super(TSNc, self).__init__(model,
                                   err_func,
                                   normalizer,
                                   beta)

    def predict(self, x, nc, significance=None):
        """Constructs forecast intervals for a set of test examples.

        Predicts the output of each test pattern using the underlying model,
        and applies the (partial) inverse nonconformity function to each
        step of the forecast, resulting in a prediction interval for each 
        step of the test pattern.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        significance : float
            Significance level (maximum allowed error rate) of predictions.
            Should be a float between 0 and 1. If ``None``, then intervals for
            all significance levels (0.01, 0.02, ..., 0.99) are output in a
            3d-matrix.

        Returns
        -------
        p : numpy array of shape [n_samples, horizon_length, 2] or [n_samples, horizon_length, 2, 99]
            If significance is ``None``, then p contains the interval (minimum
            and maximum boundaries) for each step of the test pattern, and each 
            significance level (0.01, 0.02, ..., 0.99). If significance is a 
            float between 0 and 1, then p contains the prediction intervals 
            (minimum and maximum boundaries) for the all steps of the test 
            patterns at the chosen significance level.
        """ # noqa
        n_test = x.shape[0]
        prediction = self.model.predict(x)

        norm = 1
        if self.normalizer is not None:
            try:
                norm = self.normalizer.score(x) + self.beta
            except Exception:
                pass

        if significance:
            err_dist = self.err_func.apply_inverse(nc, significance)
            err_dist = np.hstack([err_dist] * n_test)
            err_dist *= norm

            intervals = np.zeros((x.shape[0], 2))
            intervals[:, 0] = prediction - err_dist[0, :]
            intervals[:, 1] = prediction + err_dist[1, :]

            return intervals
        else:
            significance = np.arange(0.01, 1.0, 0.01)
            intervals = np.zeros((x.shape[0], self.err_func.horizon_length, 2, significance.size))

            for i, s in enumerate(significance):
                err_dist = self.err_func.apply_inverse(nc, s)
                err_dist = np.vstack([err_dist] * n_test)
                err_dist *= norm

                intervals[:, :, 0, i] = prediction - err_dist[0, :]
                intervals[:, :, 1, i] = prediction + err_dist[0, :]

            return intervals



from typing import List
import pandas as pd

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs


class Embedder(BaseEnsemble):
    """
    This ensemble acts as a simple embedder that bypasses all mixers. 
    When called, it will return the encoded representation of the data stored in (or generated by) an EncodedDs object.
    """  # noqa
    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs) -> None:
        super().__init__(target, list(), data)
        self.embedding_size = data.get_encoded_data(include_target=False).shape[-1]
        self.prepared = True

    def __call__(self, ds: EncodedDs, args: PredictionArguments = None) -> pd.DataFrame:
        # shape: (B, self.embedding_size)
        encoded_representations = ds.get_encoded_data(include_target=False).numpy()
        return pd.DataFrame(encoded_representations)



from copy import deepcopy
from typing import List, Optional

import torch
from torch import nn
from torch.optim import SGD
import numpy as np
import pandas as pd

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.stacked_ensemble import StackedEnsemble
from lightwood.encoder.array.ts_num_array import TsArrayNumericEncoder
from lightwood.api.types import PredictionArguments
from type_infer.dtype import dtype
from lightwood.data.encoded_ds import EncodedDs
from lightwood.helpers.log import log


class TsStackedEnsemble(StackedEnsemble):
    """
    Thin wrapper for `StackedEnsemble` that enables forecasting support.
    """
    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, dtype_dict: dict, ts_analysis: dict,
                 args: PredictionArguments, fit: Optional[bool] = True, **kwargs) -> None:
        dtype_dict = deepcopy(dtype_dict)
        dtype_dict[target] = dtype.float  # hijack to correctly initialize parent class
        super().__init__(target, mixers, data, dtype_dict, args, fit=False)
        if not isinstance(data.encoders[target], TsArrayNumericEncoder):
            raise Exception('This ensemble can only be used to forecast!')
        self.ts_analysis = ts_analysis
        self.horizon = self.ts_analysis['tss'].horizon
        self.target_cols = [target] + [f'{target}_timestep_{t+1}' for t in range(self.horizon - 1)]
        self.agg_dim = 2
        self.opt_max_iter = 1000

        if fit and len(mixers) > 1:
            all_preds = torch.tensor(self.predict(data, args)).squeeze().reshape(-1, self.horizon, len(mixers))
            actual = torch.tensor(data.data_frame[self.target_cols].values)
            nan_mask = actual != actual
            actual[nan_mask] = 0
            all_preds[nan_mask, :] = 0

            criterion = nn.SmoothL1Loss()
            optimizer = SGD([self.mixer_weights], lr=1e-3)

            def _eval_loss():
                optimizer.zero_grad()
                weighted = torch.sum(all_preds * self.softmax(self.mixer_weights), dim=self.agg_dim)
                loss = criterion(weighted, actual)
                loss.backward()
                return loss

            for _ in range(self.opt_max_iter):
                optimizer.step(_eval_loss)
            self.mixer_weights = self.softmax(self.mixer_weights)
            log.info(f'Optimal stacking weights: {self.mixer_weights.detach().tolist()}')
        self.prepared = True

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        assert self.prepared
        mixer_weights = torch.tensor(args.mixer_weights) if args.mixer_weights else self.mixer_weights
        output = pd.DataFrame()
        predictions = torch.tensor(np.concatenate(self.predict(ds, args), axis=2))
        predictions = (predictions * mixer_weights).sum(axis=self.agg_dim)
        output['prediction'] = predictions.detach().numpy().tolist()
        return output



import time
import inspect
from typing import Dict, List, Set, Optional
import torch
import optuna
import lightgbm
import numpy as np
import pandas as pd
from sklearn.preprocessing import OrdinalEncoder
import optuna.integration.lightgbm as optuna_lightgbm
from type_infer.dtype import dtype
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.helpers.device import get_devices
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs

optuna.logging.set_verbosity(optuna.logging.CRITICAL)


def check_gpu_support():
    try:
        data = np.random.rand(50, 2)
        label = np.random.randint(2, size=50)
        train_data = lightgbm.Dataset(data, label=label)
        params = {'num_iterations': 1, 'device': 'gpu'}
        lightgbm.train(params, train_set=train_data)
        device, nr_devices = get_devices()
        if nr_devices > 0 and str(device) != 'cpu':
            return True
        else:
            return False
    except Exception:
        return False


class LightGBM(BaseMixer):
    model: lightgbm.LGBMModel
    ordinal_encoder: OrdinalEncoder
    label_set: Set[str]
    max_bin: int
    device: torch.device
    device_str: str
    num_iterations: int
    use_optuna: bool
    supports_proba: bool

    """
    Gradient boosting mixer with a LightGBM backbone.
    
    This mixer is a good all-rounder, due to the generally great performance of tree-based ML algorithms for supervised learning tasks with tabular data.
    If you want more information regarding the techniques that set apart LightGBM from other gradient boosters, please refer to their technical paper: "LightGBM: A Highly Efficient Gradient Boosting Decision Tree" (2017).
    
    We can basically think of this mixer as a wrapper to the LightGBM interface. To do so, there are a few caveats the user may want to be aware about:
        * If you seek GPU utilization, LightGBM must be compiled from source instead of being installed through `pip`.
        * Integer, float, and quantity `dtype`s are treated as regression tasks with `L2` loss. All other supported `dtype`s is casted as a multiclass task with `multi_logloss` loss.
        * It has an automatic optuna-based hyperparameter search. This procedure triggers when a single iteration of LightGBM is deemed fast enough (given the time budget).
        * A partial fit can be performed with the `dev` data split as part of `fit`, if specified with the `fit_on_dev` argument.
    """  # noqa

    def __init__(
            self, stop_after: float, target: str, dtype_dict: Dict[str, str],
            input_cols: List[str],
            fit_on_dev: bool, use_optuna: bool, target_encoder: BaseEncoder):
        """
        :param stop_after: time budget in seconds.
        :param target: name of the target column that the mixer will learn to predict.
        :param dtype_dict: dictionary with dtypes of all columns in the data.
        :param input_cols: list of column names.
        :param fit_on_dev: whether to perform a `partial_fit()` at the end of `fit()` using the `dev` data split.
        :param use_optuna: whether to activate the automated hyperparameter search (optuna-based). Note that setting this flag to `True` does not guarantee the search will run, rather, the speed criteria will be checked first (i.e., if a single iteration is too slow with respect to the time budget, the search will not take place). 
        """  # noqa
        super().__init__(stop_after)
        self.model = None
        self.ordinal_encoder = None
        self.positive_domain = False
        self.label_set = set()
        self.target = target
        self.dtype_dict = dtype_dict
        self.input_cols = input_cols
        self.use_optuna = use_optuna
        self.params = {}
        self.fit_on_dev = fit_on_dev
        self.cls_dtypes = [dtype.categorical, dtype.binary, dtype.cat_tsarray]
        self.float_dtypes = [dtype.float, dtype.quantity, dtype.num_tsarray]
        self.num_dtypes = [dtype.integer] + self.float_dtypes
        self.supports_proba = dtype_dict[target] in self.cls_dtypes
        self.stable = True
        self.target_encoder = target_encoder

        # GPU Only available via --install-option=--gpu with opencl-dev and libboost dev (a bunch of them) installed, so let's turn this off for now and we can put it behind some flag later # noqa
        gpu_works = check_gpu_support()
        if not gpu_works:
            self.device = torch.device('cpu')
            self.device_str = 'cpu'
            log.warning(
                'LightGBM running on CPU, this somewhat slower than the GPU version, consider using a GPU instead')  # noqa
        else:
            self.device = torch.device('cuda')
            self.device_str = 'gpu'

        self.max_bin = 255

    def _to_dataset(self, data: Dict[str, Dict], output_dtype: str):
        """
        Helper method to wrangle data into the format that the underlying model requires.

        :param data: Includes train and dev data datasources.
        :param output_dtype
        :return: modified `data` object that conforms to LightGBM's expected format.
        """
        weight_map = getattr(self.target_encoder, 'target_weights', None)

        for subset_name in data.keys():
            for input_col in self.input_cols:
                if data[subset_name]['data'] is None:
                    data[subset_name]['data'] = data[subset_name]['ds'].get_encoded_column_data(
                        input_col).to(self.device)
                else:
                    enc_col = data[subset_name]['ds'].get_encoded_column_data(input_col)
                    data[subset_name]['data'] = torch.cat((data[subset_name]['data'], enc_col.to(self.device)), 1)

            data[subset_name]['data'] = data[subset_name]['data'].cpu().numpy()

            label_data = data[subset_name]['ds'].get_column_original_data(self.target)

            data[subset_name]['weights'] = None
            if output_dtype in self.cls_dtypes:
                if subset_name == 'train':
                    self.ordinal_encoder = OrdinalEncoder()
                    self.label_set = set(label_data)
                    self.label_set.add('__mdb_unknown_cat')
                    self.ordinal_encoder.fit(np.array(list(self.label_set)).reshape(-1, 1))

                label_data = [x if x in self.label_set else '__mdb_unknown_cat' for x in label_data]
                if weight_map is not None:
                    data[subset_name]['weights'] = [weight_map[x] for x in label_data]
                label_data = self.ordinal_encoder.transform(np.array(label_data).reshape(-1, 1)).flatten()
            elif output_dtype in self.num_dtypes:
                if weight_map is not None:
                    target_encoder = data[subset_name]['ds'].encoders[self.target]

                    # get the weights from the numeric target encoder
                    data[subset_name]['weights'] = target_encoder.get_weights(label_data)

                if output_dtype in self.float_dtypes:
                    label_data = label_data.astype(float)
                elif output_dtype == dtype.integer:
                    label_data = label_data.clip(-pow(2, 63), pow(2, 63)).astype(int)

            data[subset_name]['label_data'] = label_data

        return data

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the LightGBM model.

        :param train_data: encoded features for training dataset
        :param dev_data: encoded features for dev dataset
        """
        started = time.time()

        log.info('Started fitting LGBM model')
        data = {
            'train': {'ds': train_data, 'data': None, 'label_data': {}},
            'dev': {'ds': dev_data, 'data': None, 'label_data': {}}
        }
        self.fit_data_len = len(data['train']['ds'])
        self.positive_domain = getattr(train_data.encoders.get(self.target, None), 'positive_domain', False)

        output_dtype = self.dtype_dict[self.target]
        data = self._to_dataset(data, output_dtype)

        if output_dtype not in self.cls_dtypes + self.num_dtypes:
            log.error(f'Lightgbm mixer not supported for type: {output_dtype}')
            raise Exception(f'Lightgbm mixer not supported for type: {output_dtype}')
        else:
            objective = 'regression' if output_dtype in self.num_dtypes else 'multiclass'
            metric = 'l2' if output_dtype in self.num_dtypes else 'multi_logloss'

        self.params = {
            'objective': objective,
            'metric': metric,
            'verbose': -1,
            'lambda_l1': 0.1,
            'lambda_l2': 0.1,
            'force_row_wise': True,
            'device_type': self.device_str,
        }

        if objective == 'multiclass':
            self.all_classes = self.ordinal_encoder.categories_[0]
            self.params['num_class'] = self.all_classes.size
        elif output_dtype == dtype.num_tsarray:
            self.params['linear_tree'] = True
        if self.device_str == 'gpu':
            self.params['gpu_use_dp'] = True

        # Determine time per iterations
        start = time.time()
        self.params['num_iterations'] = 1
        '''
        Why construst a dataset here instead of using the training dataset?

        Because it guards against the following crash:

        WARNING:lightwood-1613058:Exception: Cannot change feature_pre_filter after constructed Dataset handle. when training mixer: <lightwood.mixer.lightgbm.LightGBM object at 0x7f39148eae20>
        feature_fraction, val_score: inf:   0%|                                 | 0/7 [00:00<?, ?it/s]
        free(): double free detected in tcache 2

        Only happens sometimes and I can find no pattern as to when, happens for multiple input and target types.

        Why does the following crash happen and what does it mean? No idea, closest relationships I can find is /w optuna modifying parameters after the dataset is create: https://github.com/microsoft/LightGBM/issues/4019 | But why this would apply here makes no sense. Could have to do with the `train` process of lightgbm itself setting a "set only once" property on a dataset when it starts. Dunno, if you find out replace this comment with the real reason.
        '''  # noqa
        kwargs = {}
        if 'verbose_eval' in inspect.getfullargspec(lightgbm.train).args:
            kwargs['verbose_eval'] = False
        self.model = lightgbm.train(self.params,
                                    lightgbm.Dataset(data['train']['data'],
                                                     label=data['train']['label_data'],
                                                     weight=data['train']['weights']),
                                    **kwargs)
        end = time.time()
        seconds_for_one_iteration = max(0.1, end - start)

        self.stop_after = max(1, self.stop_after - (time.time() - started))
        # Determine nr of iterations
        log.info(f'A single GBM iteration takes {seconds_for_one_iteration} seconds')
        self.num_iterations = int(self.stop_after * 0.8 / seconds_for_one_iteration)

        # Turn on grid search if training doesn't take too long using it
        kwargs = {}
        if self.use_optuna and self.num_iterations >= 200:
            model_generator = optuna_lightgbm
            kwargs['time_budget'] = self.stop_after * 0.4
            self.num_iterations = int(self.num_iterations / 2)
            kwargs['optuna_seed'] = 0
        else:
            model_generator = lightgbm

        # Train the models
        log.info(
            f'Training GBM ({model_generator}) with {self.num_iterations} iterations given {self.stop_after} seconds constraint')  # noqa
        if self.num_iterations < 1:
            self.num_iterations = 1
        self.params['num_iterations'] = int(self.num_iterations)

        self.params['early_stopping_rounds'] = 5

        # Prepare the data
        train_dataset = lightgbm.Dataset(data['train']['data'], label=data['train']['label_data'],
                                         weight=data['train']['weights'])

        dev_dataset = lightgbm.Dataset(data['dev']['data'], label=data['dev']['label_data'],
                                       weight=data['dev']['weights'])

        if 'verbose_eval' in inspect.getfullargspec(lightgbm.train).args:
            kwargs['verbose_eval'] = False
        self.model = model_generator.train(
            self.params, train_dataset, valid_sets=[dev_dataset, train_dataset],
            valid_names=['dev', 'train'],
            **kwargs)
        self.num_iterations = self.model.best_iteration
        log.info(f'Lightgbm model contains {self.model.num_trees()} weak estimators')

        if self.fit_on_dev:
            self.partial_fit(dev_data, train_data)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        """
        Updates the LightGBM model.

        :param train_data: encoded features for (new) training dataset
        :param dev_data: encoded features for (new) dev dataset
        """
        pct_of_original = len(train_data) / self.fit_data_len
        iterations = max(1, int(self.num_iterations * pct_of_original) / 2)

        data = {'retrain': {'ds': train_data, 'data': None, 'label_data': {}}, 'dev': {
            'ds': dev_data, 'data': None, 'label_data': {}}}

        output_dtype = self.dtype_dict[self.target]
        data = self._to_dataset(data, output_dtype)

        train_dataset = lightgbm.Dataset(data['retrain']['data'], label=data['retrain']['label_data'],
                                         weight=data['retrain']['weights'])
        dev_dataset = lightgbm.Dataset(data['dev']['data'], label=data['dev']['label_data'],
                                       weight=data['dev']['weights'])

        log.info(f'Updating lightgbm model with {iterations} iterations')
        self.params['num_iterations'] = int(iterations)

        kwargs = {}
        if 'verbose_eval' in inspect.getfullargspec(lightgbm.train).args:
            kwargs['verbose_eval'] = False

        self.model = lightgbm.train(
            self.params, train_dataset, valid_sets=[dev_dataset, train_dataset],
            valid_names=['dev', 'retrain'],
            init_model=self.model, **kwargs)
        log.info(f'Model now has a total of {self.model.num_trees()} weak estimators')

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Call a trained LightGBM mixer to output predictions for the target column.

        :param ds: input data with values for all non-target columns.
        :param args: inference-time arguments (e.g. whether to output predicted labels or probabilities).

        :return: dataframe with predictions.
        """
        data = None
        for input_col in self.input_cols:
            if data is None:
                data = ds.get_encoded_column_data(input_col).to(self.device)
            else:
                data = torch.cat((data, ds.get_encoded_column_data(input_col).to(self.device)), 1)

        data = data.cpu().numpy()
        raw_predictions = self.model.predict(data)

        if self.ordinal_encoder is not None:
            decoded_predictions = self.ordinal_encoder.inverse_transform(
                np.argmax(raw_predictions, axis=1).reshape(-1, 1)).flatten()
        else:
            decoded_predictions = raw_predictions

        if self.positive_domain:
            decoded_predictions = [max(0, p) for p in decoded_predictions]

        ydf = pd.DataFrame({'prediction': decoded_predictions})

        if args.predict_proba and self.ordinal_encoder is not None:
            for idx, label in enumerate(self.ordinal_encoder.categories_[0].tolist()):
                ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

        return ydf



from typing import Dict, Optional, Union

import pandas as pd

from lightwood.mixer.sktime import SkTime


class ProphetMixer(SkTime):
    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            horizon: int,
            ts_analysis: Dict,
            auto_size: bool = True,
            use_stl: bool = False,
            add_seasonality: Optional[dict] = None,
            add_country_holidays: Optional[dict] = None,
            growth: str = 'linear',
            growth_floor: float = 0,
            growth_cap: Optional[float] = None,
            changepoints: Optional[list] = None,
            n_changepoints: int = 25,
            changepoint_range: float = 0.8,
            yearly_seasonality: Union[str, bool, int] = 'auto',
            weekly_seasonality: Union[str, bool, int] = 'auto',
            daily_seasonality: Union[str, bool, int] = 'auto',
            holidays: Optional[pd.DataFrame] = None,
            seasonality_mode: str = 'additive',
            seasonality_prior_scale: float = 10.0,
            holidays_prior_scale: float = 10.0,
            changepoint_prior_scale: float = 0.05,
            mcmc_samples: int = 0,
            alpha: float = 0.05,
            uncertainty_samples: int = 1000,
    ):
        """
        Wrapper for SkTime's Prophet interface.
         
        :param stop_after: time budget in seconds
        :param target: column containing target time series
        :param dtype_dict: data types for each dataset column
        :param horizon: forecast length
        :param ts_analysis: lightwood-produced stats about input time series
        :param auto_size: whether to filter out old data points if training split is bigger than a certain threshold (defined by the dataset sampling frequency). Enabled by default to avoid long training times in big datasets.
        :param use_stl: Whether to use de-trenders and de-seasonalizers fitted in the timeseries analysis phase.
        
        For the rest of the parameters, please refer to SkTime's documentation.
        """  # noqa

        # overrides
        sp = 1
        hyperparam_search = False
        model_path = 'fbprophet.Prophet'
        model_kwargs = {
            'add_seasonality': add_seasonality,
            'add_country_holidays': add_country_holidays,
            'growth': growth,
            'growth_floor': growth_floor,
            'growth_cap': growth_cap,
            'changepoints': changepoints,
            'n_changepoints': n_changepoints,
            'changepoint_range': changepoint_range,
            'yearly_seasonality': yearly_seasonality,
            'weekly_seasonality': weekly_seasonality,
            'daily_seasonality': daily_seasonality,
            'holidays': holidays,
            'seasonality_mode': seasonality_mode,
            'seasonality_prior_scale': seasonality_prior_scale,
            'holidays_prior_scale': holidays_prior_scale,
            'changepoint_prior_scale': changepoint_prior_scale,
            'mcmc_samples': mcmc_samples,
            'alpha': alpha,
            'uncertainty_samples': uncertainty_samples,
        }

        # setup sktime base mixer
        super().__init__(stop_after, target, dtype_dict, horizon, ts_analysis,
                         model_path, model_kwargs, auto_size, sp, hyperparam_search, use_stl)
        self.name = 'Prophet'
        self.stable = False



from typing import Union, List, Dict

import torch
import numpy as np
import pandas as pd

from lightwood.encoder.numeric import NumericEncoder


class TsNumericEncoder(NumericEncoder):
    """
    Variant of vanilla numerical encoder, supports dynamic mean re-scaling
    """
    is_timeseries_encoder: bool = True

    def __init__(self, is_target: bool = False, positive_domain: bool = False, grouped_by=None):
        super(TsNumericEncoder, self).__init__(is_target=is_target, positive_domain=positive_domain)
        # time series normalization params
        self.normalizers = None
        self.group_combinations = None
        self.dependencies = grouped_by
        self.output_size = 1

    def encode(self, data: Union[np.ndarray, pd.Series], dependency_data: Dict[str, List[pd.Series]] = {}):
        """
        :param data: A pandas series containing the numbers to be encoded
        :param dependency_data: dict with grouped_by column info, to retrieve the correct normalizer for each datum

        :returns: A torch tensor with the representations of each number
        """  # noqa
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if not dependency_data:
            dependency_data = {'__default': [None] * len(data)}

        if isinstance(data, pd.Series):
            data = data.values

        # get array of series-wise observed means
        if self.normalizers is None:
            means = np.full((len(data)), fill_value=self._abs_mean)
        else:
            # use global mean as default for novel series
            means = np.full((len(data)), fill_value=self.normalizers['__default'].abs_mean)

            def _get_group_mean(group) -> float:
                if (group, ) in self.normalizers:
                    return self.normalizers[(group, )].abs_mean
                else:
                    return self.normalizers['__default'].abs_mean

            for i, group in enumerate(list(zip(*dependency_data.values()))):  # TODO: support multigroup
                if group[0] is not None:
                    means = np.vectorize(_get_group_mean, otypes=[float])(group[0].values)

        if len(data.shape) > 1 and data.shape[1] > 1:
            if len(means.shape) == 1:
                means = np.expand_dims(means, 1)
            means = np.repeat(means, data.shape[1], axis=1)

        def _norm_fn(x: float, mean: float) -> float:
            return x / mean

        # nones = np.vectorize(self._none_fn, otypes=[float])(data)  # TODO
        encoded = np.vectorize(_norm_fn, otypes=[float])(data, means)
        # encoded[nones] = 0  # if measurement is None, it is zeroed out  # TODO

        # TODO: mask for where mean is 0, then pass real as-is

        return torch.Tensor(encoded).unsqueeze(1)

    def decode(self, encoded_values: torch.Tensor, decode_log: bool = None, dependency_data=None):
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        assert isinstance(encoded_values, torch.Tensor), 'It is not a tensor!'  # TODO: debug purposes
        assert not decode_log  # TODO: debug purposes

        if not dependency_data:
            dependency_data = {'__default': [None] * len(encoded_values)}

        # force = True prevents side effects on the original encoded_values
        ev = encoded_values.numpy(force=True)

        # set global mean as default
        ret = np.full((ev.shape[0],), dtype=float, fill_value=self._abs_mean)

        # TODO: perhaps capture nan, infs, etc and set to pow(10,63)?

        # set means array
        if self.normalizers is None:
            means = np.full((ev.shape[0],), fill_value=self._abs_mean)
        else:
            means = np.full((len(encoded_values)), fill_value=self.normalizers['__default'].abs_mean)
            for i, group in enumerate(list(zip(*dependency_data.values()))):
                if group is not None:
                    if tuple(group) in self.normalizers:
                        means[i] = self.normalizers[tuple(group)].abs_mean
                    else:
                        means[i] = self.normalizers['__default'].abs_mean
                else:
                    means[i] = self._abs_mean

        # set real value
        real_value = np.multiply(ev[:].reshape(-1,), means)
        valid_mask = np.ones_like(real_value, dtype=bool)

        # final filters
        if self.positive_domain:
            real_value = abs(real_value)

        ret[valid_mask] = real_value[valid_mask]  # TODO probably not needed

        return ret.tolist()



from typing import Dict, Union, Optional
from copy import deepcopy

import numpy as np
import pandas as pd
from neuralforecast import NeuralForecast
from neuralforecast.models.nhits import NHITS
from neuralforecast.models.nbeats import NBEATS
from neuralforecast.losses.pytorch import MQLoss

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class NHitsMixer(BaseMixer):
    horizon: int
    target: str
    supports_proba: bool
    model_path: str
    hyperparam_search: bool
    default_config: dict
    SUPPORTED_MODELS = ('nhits', 'nbeats')

    def __init__(
            self,
            stop_after: float,
            target: str,
            horizon: int,
            window: int,
            dtype_dict: Dict,
            ts_analysis: Dict,
            pretrained: bool = False,
            train_args: Optional[Dict] = None,
    ):
        """
        Wrapper around an N-HITS deep learning model.
        
        :param stop_after: time budget in seconds.
        :param target: column to forecast.
        :param horizon: length of forecasted horizon.
        :param window: length of input data.
        :param ts_analysis: dictionary with miscellaneous time series info, as generated by 'lightwood.data.timeseries_analyzer'.
        :param train_args: arguments to steer the training process. 
            - `trainer_args`: all arguments for the PyTorchLightning trainer.
            - `conf_level`: level passed into MQLoss. Directly impacts prediction bounds. 
        """  # noqa
        super().__init__(stop_after)
        self.stable = False
        self.prepared = False
        self.supports_proba = False
        self.target = target
        self.window = window
        self.horizon = horizon
        self.dtype_dict = dtype_dict
        self.ts_analysis = ts_analysis
        self.grouped_by = ['__default'] if not ts_analysis['tss'].group_by else ts_analysis['tss'].group_by
        self.group_boundaries = {}  # stores last observed timestamp per series
        self.train_args = train_args.get('trainer_args', {}) if train_args else {}

        # we set a fairly aggressive training schedule by default
        self.train_args['early_stop_patience_steps'] = self.train_args.get('early_stop_patience_steps', 1)
        self.train_args['val_check_steps'] = self.train_args.get('val_check_steps', 10)
        self.train_args['learning_rate'] = self.train_args.get('learning_rate', 3e-3)
        self.train_args['mlp_units'] = self.train_args.get('mlp_units', [[128, 128], [128, 128]])
        self.train_args['random_seed'] = self.train_args.get('random_seed', 1)

        self.conf_level = self.train_args.pop('conf_level', [90])
        for level in self.conf_level:
            assert 0 <= level <= 100, f'A provided level is not in the [0, 100] range (found: {level})'
            assert isinstance(level, int), f'A provided level is not an integer (found: {level})'

        self.pretrained = pretrained
        self.base_url = 'https://nixtla-public.s3.amazonaws.com/transfer/pretrained_models/'
        self.freq_to_model = {
            'Y': 'yearly',
            'Q': 'monthly',
            'M': 'monthly',
            'W': 'daily',
            'D': 'daily',
            'H': 'hourly',
            'T': 'hourly',  # NOTE: use another pre-trained model once available
            'S': 'hourly'  # NOTE: use another pre-trained model once available
        }
        self.model = None
        self.model_class_str = self.train_args.get('model_class', 'nhits').lower()
        assert self.model_class_str in NHitsMixer.SUPPORTED_MODELS, f'Provided model class ({self.model_class_str}) is not supported. Supported models are: {NHitsMixer.SUPPORTED_MODELS}'  # noqa
        self.model_class = NBEATS if self.model_class_str == 'nbeats' else NHITS
        self.model_name = None
        self.model_names = {
            'nhits': {
                'hourly': 'nhits_m4_hourly.ckpt',  # hourly (non-tiny)
                'daily': 'nhits_m4_daily.ckpt',   # daily
                'monthly': 'nhits_m4_monthly.ckpt',  # monthly
                'yearly': 'nhits_m4_yearly.ckpt',  # yearly
            },
            'nbeats': {}  # TODO: complete
        }

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits the NeuralForecast model.
        """  # noqa
        log.info('Started fitting N-HITS forecasting model')

        # prepare data
        cat_ds = ConcatedEncodedDs([train_data, dev_data])
        oby_col = self.ts_analysis["tss"].order_by
        gby = self.ts_analysis["tss"].group_by if self.ts_analysis["tss"].group_by else []
        df = deepcopy(cat_ds.data_frame)
        Y_df, _ = self._make_initial_df(df, mode='train')
        self.group_boundaries = self._set_boundary(Y_df, gby)
        if gby:
            n_time = df[gby].value_counts().min()
        else:
            n_time = len(df[f'__mdb_original_{oby_col}'].unique())
        n_ts_val = max(int(.1 * n_time), self.horizon)  # at least self.horizon to validate on

        # train the model
        n_time_out = self.horizon
        if self.pretrained:
            # TODO: let user specify finetuning
            self.model_name = self.model_names.get(self.freq_to_model[self.ts_analysis['sample_freqs']['__default']],
                                                   None)
            self.model_name = self.model_names['hourly'] if self.model_name is None else self.model_name
            ckpt_url = self.base_url + self.model_name
            self.model = self.model_class.load_from_checkpoint(ckpt_url)

            if not self.window < self.model.hparams.n_time_in:
                log.info(f'NOTE: Provided window ({self.window}) is smaller than specified model input length ({self.model.hparams.n_time_in}). Will train a new model from scratch.')  # noqa
                self.pretrained = False
            if self.horizon > self.model.hparams.n_time_out:
                log.info(f'NOTE: Horizon ({self.horizon}) is bigger than that of the pretrained model ({self.model.hparams.n_time_out}). Will train a new model from scratch.')  # noqa
                self.pretrained = False
            if self.pretrained:
                log.info(f'Successfully loaded pretrained N-HITS forecasting model ({self.model_name})')

        if not self.pretrained:
            if self.window + self.horizon > n_time:
                new_window = max(1, n_time - self.horizon - 1)
                self.window = new_window
                log.info(f'Window {self.window} is too long for data provided (group: {df[gby].value_counts()[::-1].index[0]}), reducing window to {new_window}.')  # noqa
            model = self.model_class(h=n_time_out, input_size=self.window, **self.train_args, loss=MQLoss(level=self.conf_level))  # noqa
            self.model = NeuralForecast(models=[model], freq=self.ts_analysis['sample_freqs']['__default'],)
            self.model.fit(df=Y_df, val_size=n_ts_val)
            log.info('Successfully trained N-HITS forecasting model.')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        self.hyperparam_search = False
        self.train_args = args.get('trainer_args', {}) if args else {}  # NOTE: this replaces the original config
        self.fit(train_data, dev_data)
        self.prepared = True

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Calls the mixer to emit forecasts.
        
        NOTE: in the future we may support predicting every single row efficiently. For now, this mixer
        replicates the neuralforecast library behavior and returns a forecast strictly for the next `tss.horizon`
        timesteps after the end of the input dataframe.
        """  # noqa
        if args.predict_proba:
            log.warning('This mixer does not output probability estimates')

        # provided quantile must match one of the training levels, else we default to the largest one of these
        if args.fixed_confidence is not None and int(args.fixed_confidence * 100) in self.conf_level:
            level = int(args.fixed_confidence * 100)
        else:
            level = max(self.conf_level)

        target_cols = ['prediction', 'lower', 'upper']
        pred_cols = [
            f'{self.model_class_str.upper()}-median',
            f'{self.model_class_str.upper()}-lo-{level}',
            f'{self.model_class_str.upper()}-hi-{level}'
        ]

        input_df, idxs = self._make_initial_df(deepcopy(ds.data_frame))
        length = sum(ds.encoded_ds_lengths) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0, index=range(length), columns=target_cols, dtype=object)

        # fill with zeroed arrays
        zero_array = [0 for _ in range(self.horizon)]
        for target_col in target_cols:
            ydf[target_col] = [zero_array] * len(ydf)

        grouper = input_df.groupby('unique_id')
        group_ends = grouper.last()['index'].values
        fcst = self.model.predict(input_df).reset_index()
        fcst['ds'] = fcst.groupby('unique_id').cumcount()
        horizons = pd.pivot_table(fcst, values=pred_cols, index='unique_id', columns='ds')

        temp_df = pd.DataFrame(0,  # zero-filled
                               index=range(len(horizons)),
                               columns=target_cols,
                               dtype=object)

        for pcol, tcol in zip(pred_cols, target_cols):
            temp_df[tcol] = horizons[pcol].values.tolist()

        for tcol in target_cols:
            ydf[tcol].iloc[group_ends] = temp_df[tcol]

        ydf['confidence'] = level / 100
        return ydf

    def _make_initial_df(self, df, mode='inference'):
        """
        Prepares a dataframe for the model according to what neuralforecast expects.

        If a per-group boundary exists, this method additionally drops out all observations prior to the cutoff.
        """  # noqa

        oby_col = self.ts_analysis["tss"].order_by
        # df = df.sort_values(by=f'__mdb_original_{oby_col}')  # TODO rm
        df[f'__mdb_parsed_{oby_col}'] = df.index
        df = df.reset_index(drop=True)

        Y_df = pd.DataFrame()
        Y_df['_index'] = np.arange(len(df))
        Y_df['y'] = df[self.target]
        Y_df['ds'] = df[f'__mdb_parsed_{oby_col}']

        if self.grouped_by != ['__default']:
            Y_df['unique_id'] = df[self.grouped_by].apply(lambda x: ','.join([elt for elt in x]), axis=1)
        else:
            Y_df['unique_id'] = '__default'

        Y_df = Y_df.reset_index()

        # filter if boundary exists
        if mode == 'train' and self.group_boundaries:
            filtered = []
            grouped = Y_df.groupby(by='unique_id')
            for group, sdf in grouped:
                if group in self.group_boundaries:
                    sdf = sdf[sdf['ds'].gt(self.group_boundaries[group])]
                    if sdf.shape[0] > 0:
                        filtered.append(sdf)
            if filtered:
                Y_df = pd.concat(filtered)

        filtered_idxs = Y_df.pop('_index').values
        return Y_df, filtered_idxs

    @staticmethod
    def _set_boundary(df: pd.DataFrame, gby: list) -> Dict[str, object]:
        """
        Finds last observation for every series in a pre-sorted `df` given a `gby` list of columns to group by.
        """
        if not gby:
            group_boundaries = {'__default': df.iloc[-1]['ds']}
        else:
            # could use groupby().transform('max'), but we leverage pre-sorting instead
            grouped_df = df.groupby(by='unique_id', as_index=False).last()
            group_boundaries = grouped_df[['unique_id', 'ds']].set_index('unique_id').to_dict()['ds']

        return group_boundaries



import functools
import torch
from torch.nn.functional import pad
from lightwood.helpers.device import get_devices


def concat_vectors_and_pad(vec_list, max_):
    """
    Concatenates a list of input vectors and pads them to match a specified maximum
    length.

    This function takes a list of input vectors, concatenates them along a specified
    dimension (dim=0), and then pads the concatenated vector to achieve a specified
    maximum length. The padding is done with zeros.

    Args:
        vec_list (list of torch.Tensor): List of input vectors to concatenate and pad.
        max_ (int): The maximum length of the concatenated and padded vector.

    Returns:
        torch.Tensor: The concatenated and padded vector.

    Raises:
        AssertionError: If the length of 'vec_list' is not greater than 0, or if it
                        exceeds 'max_len', or if 'max_len' is not greater than 0.

    Example:
        >>> input_tensors = [torch.tensor([1, 2]), torch.tensor([3, 4, 5])]
        >>> max_length = 5
        >>> concatenated_padded = concat_vectors_and_pad(input_tensors, max_length)
        >>> print(concatenated_padded)
        tensor([1, 2, 3, 4, 5])
    """
    assert len(vec_list) > 0
    assert len(vec_list) <= max_
    assert max_ > 0

    cat_vec = torch.cat(list(vec_list), dim=0)

    pad_size = max_ - len(vec_list)
    padding = (0, pad_size * vec_list[0].size(0))
    padded = pad(cat_vec[None], padding, 'constant', 0)[0]

    return padded


def average_vectors(vec_list):
    assert len(vec_list) > 0
    return torch.cat([emb[None] for emb in vec_list], dim=0).mean(0)


class LightwoodAutocast:
    """
    Equivalent to torch.cuda.amp.autocast, but checks device compute capability
    to activate the feature only when the GPU has tensor cores to leverage AMP.

    **Attributes:**

    * `active` (bool): Whether AMP is currently active. This attribute is at the class
    level

    **Usage:**

    ```python
    >>> import lightwood.helpers.torch as lt
    >>> with lt.LightwoodAutocast():
    ...     # This code will be executed in AMP mode.
    ...     pass
    """
    active = False

    def __init__(self, enabled=True):
        """
        Initializes the context manager for Automatic Mixed Precision (AMP) functionality.

        Args:
            enabled (bool, optional): Whether to enable AMP. Defaults to True.
        """
        self.major = 0  # GPU major version
        torch_version = [int(i) for i in torch.__version__.split('.')[:-1]]

        if not enabled or not torch.cuda.is_available() or torch_version[0] < 1 or torch_version[1] < 6:
            self._enabled = False
        else:
            device, _ = get_devices()
            if device.type == 'cuda':
                # tensor cores only exist from 7 onwards
                # if this is not the case, then AMP is unnecessary overhead
                self.major, _ = torch.cuda.get_device_capability(device)
                self._enabled = enabled if self.major > 6 else False
            else:
                self._enabled = False  # gpu is available but cpu is forced

        self.prev = self._enabled  # necessary reference to exit
        LightwoodAutocast.active = self._enabled

    def __enter__(self):
        """
        * `__enter__()`: Enters the context manager and enables AMP if it is not already enabled.
        """
        if self._enabled:
            self.prev = torch.is_autocast_enabled()
            torch.set_autocast_enabled(self._enabled)
            torch.autocast_increment_nesting()

    def __exit__(self, *args):
        """
        * `__exit__()`: Exits the context manager and disables AMP.
        """
        if self._enabled:
            # Drop the cache when we exit to a nesting level that's outside any instance of autocast
            if torch.autocast_decrement_nesting() == 0:
                torch.clear_autocast_cache()
            torch.set_autocast_enabled(self.prev)
        return False

    def __call__(self, func):
        """
        * `__call__(self, func)`: Returns a decorated function that enables AMP when it is called.
        """
        @functools.wraps(func)
        def decorate_autocast(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorate_autocast



from typing import List, Optional

import numpy as np
import pandas as pd
from mindsdb_evaluator import evaluate_accuracies

from lightwood.helpers.log import log
from type_infer.helpers import is_nan_numeric
from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments, SubmodelData
from lightwood.data.encoded_ds import EncodedDs

# special dispatches
from lightwood.mixer import GluonTSMixer  # imported from base mixer folder as it needs optional dependencies


class BestOf(BaseEnsemble):
    """
    This ensemble acts as a mixer selector. 
    After evaluating accuracy for all internal mixers with the validation data, it sets the best mixer as the underlying model.
    """  # noqa
    indexes_by_accuracy: List[float]

    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, accuracy_functions,
                 args: PredictionArguments, ts_analysis: Optional[dict] = None, fit: bool = True) -> None:
        super().__init__(target, mixers, data, fit=False)
        self.special_dispatch_list = [GluonTSMixer]

        if fit:
            score_list = []
            for _, mixer in enumerate(self.mixers):

                if type(mixer) in self.special_dispatch_list:
                    avg_score = self.special_dispatch(mixer, data, args, target, ts_analysis)
                else:
                    score_dict = evaluate_accuracies(
                        data.data_frame,
                        mixer(data, args)['prediction'],
                        target,
                        accuracy_functions,
                        ts_analysis=ts_analysis
                    )

                    avg_score = np.mean(list(score_dict.values()))
                log.info(f'Mixer: {type(mixer).__name__} got accuracy: {avg_score}')

                if is_nan_numeric(avg_score):
                    avg_score = -pow(2, 63)
                    log.warning(f'Change the accuracy of mixer {type(mixer).__name__} to valid value: {avg_score}')

                self.submodel_data.append(SubmodelData(
                    name=type(mixer).__name__,
                    accuracy=avg_score,
                    is_best=False
                ))

                score_list.append(avg_score)

            self.indexes_by_accuracy = list(reversed(np.array(score_list).argsort()))
            self.supports_proba = self.mixers[self.indexes_by_accuracy[0]].supports_proba
            log.info(f'Picked best mixer: {type(self.mixers[self.indexes_by_accuracy[0]]).__name__}')
            self.prepared = True
            self.submodel_data[self.indexes_by_accuracy[0]].is_best = True

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        assert self.prepared
        if args.all_mixers:
            predictions = {}
            for mixer in self.mixers:
                predictions[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']
            return pd.DataFrame(predictions)
        else:
            for mixer_index in self.indexes_by_accuracy:
                mixer = self.mixers[mixer_index]
                try:
                    return mixer(ds, args=args)
                except Exception as e:
                    if mixer.stable:
                        raise(e)
                    else:
                        log.warning(f'Unstable mixer {type(mixer).__name__} failed with exception: {e}.\
                        Trying next best')

    def special_dispatch(self, mixer, data, args, target, ts_analysis):
        if isinstance(mixer, GluonTSMixer):
            return mixer.model_train_stats.loss_history[-1]



import random
import torch
import numpy as np

try:
    import mxnet as mx
except Exception:
    mx = None


def seed(seed_nr: int) -> None:
    torch.manual_seed(seed_nr)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    np.random.seed(seed_nr)
    random.seed(seed_nr)

    if mx is not None:
        mx.random.seed(seed_nr)



from typing import List
import torch
from torch import nn
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.helpers.device import get_device_from_name, get_devices
from lightwood.helpers.log import log


class ResidualModule(nn.Module):
    def __init__(
        self,
        input_size
    ) -> None:
        """Initialize self."""
        intermediate_size = max([input_size * 2, 400])
        super().__init__()
        self.normalization = nn.BatchNorm1d(input_size)
        self.linear_first = nn.Linear(input_size, intermediate_size)
        self.activation_first = nn.SELU()
        self.linear_second = nn.Linear(intermediate_size, input_size)
        self.dropout = nn.Dropout(0.1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Perform the forward pass."""
        x_input = x
        if self.training:
            x = self.normalization(x)
        x = self.linear_first(x)
        x = self.activation_first(x)
        x = self.dropout(x)
        x = self.linear_second(x)
        x = x_input + x
        return x


class ResidualNet(torch.nn.Module):
    def __init__(self,
                 input_size: int = None,
                 output_size: int = None,
                 shape: List[int] = None,
                 max_params: int = int(3e5),
                 device: str = '') -> None:
        super(ResidualNet, self).__init__()
        self.net = torch.nn.Sequential(
            *
            ([ResidualModule(input_size) for _ in range(1)] +
             [nn.Linear(input_size, max([input_size * 2, output_size * 2, 400])),
              nn.Linear(max([input_size * 2, output_size * 2, 400]),
                        output_size)]))

        if(device == ''):
            device, available_devices = get_devices()
        else:
            device = get_device_from_name(device)
            available_devices = 0
        self.to(device, available_devices)

    def to(self, device: torch.device, available_devices: int = 1) -> torch.nn.Module:
        self.net = self.net.to(device)
        if available_devices > 1:
            self.dp_wrapper_net = torch.nn.DataParallel(self.net)
        else:
            self.dp_wrapper_net = self.net

        self.device = device
        self.available_devices = available_devices
        return self

    def forward(self, input):
        try:
            with LightwoodAutocast():
                output = self.net(input)
        except Exception as e:
            # Data parallel error
            if 'nccl' in str(e).lower():
                self.dp_wrapper_net = self.net
                log.warn(f'Data parallel not working: {e}')
            else:
                raise e

        return output



from typing import Callable
from collections import deque

import inspect
import numpy as np

from type_infer.dtype import dtype


def is_allowed(v):
    if '(' in str(v):
        return False
    if 'lambda' in str(v):
        return False
    if '__' in str(v):
        return False

    return True


def call(entity: dict) -> str:
    # Special behavior for ensemble
    if 'submodels' in entity['args']:
        del entity['args']['submodels']

    for k, v in entity['args'].items():
        if not str(v).startswith('$'):
            if not is_allowed(v):
                raise Exception(f'Invalid value: {v} for arg {k}')

    args = [f'{k}={v}' for k, v in entity['args'].items() if not str(v).startswith('$')]

    for k, v in entity['args'].items():
        if str(v).startswith('$'):
            v = str(v).replace('$', 'self.')
            args.append(f'{k}={v}')

    args = ','.join(args)
    return f"""{entity['module']}({args})"""


def inline_dict(obj: dict) -> str:
    arr = []
    for k, v in obj.items():
        if str(v) in list(dtype.__dict__.keys()):
            v = f"'{v}'"
        k = k.replace("'", "\\'").replace('"', '\\"')
        arr.append(f"""'{k}': {v}""")

    dict_code = '{\n' + ',\n'.join(arr) + '\n}'
    return dict_code


def align(code: str, indent: int) -> str:
    add_space = ''
    for _ in range(indent):
        add_space += '    '

    code_arr = code.split('\n')
    code = f'\n{add_space}'.join(code_arr)
    return code


def _consolidate_analysis_blocks(jsonai, key):
    """
    Receives a list of analysis blocks (where applicable, already filed with `hidden` args) and modifies it so that:
        1. All dependencies are correct.
        2. Execution order is such that all dependencies are met.
            - For this we use a topological sort over the DAG.
    """
    analysis_defaults = {  # non-optional plus dependencies  # TODO: move to config file
        'ICP': {
            "deps": [],
        },
        'AccStats': {
            "deps": ['ICP']
        },
        'ConfStats': {
            "deps": ['ICP']
        },
        'PermutationFeatureImportance': {
            "deps": ['AccStats']
        },
        'ShapleyValues': {
            "deps": []
        },
        'TempScaler': {
            "deps": []
        }
    }

    # 1. all dependencies are correct
    blocks = getattr(jsonai, key)
    block_objs = {b['module']: b for b in blocks}

    for i, block in enumerate(blocks):
        if 'args' not in block:
            blocks[i]['args'] = analysis_defaults.get(block['module'], {"deps": []})
        elif 'deps' not in block['args']:
            blocks[i]['args']['deps'] = []
        for dep in block['args']['deps']:
            if dep not in block_objs.keys():
                raise Exception(f'Analysis block "{dep}" not found but necessary for block "{block["module"]}". Please add it and try again.')  # noqa

    # 2. correct execution order -- build a DAG out of analysis blocks
    block_objs = {b['module']: b for b in blocks}
    block_ids = {k: i for i, k in enumerate(block_objs.keys())}
    idx2block = {i: k for i, k in enumerate(block_objs.keys())}

    adj_M = np.zeros((len(block_ids), len(block_ids)))
    for k, b in block_objs.items():
        for dep in b['args']['deps']:
            adj_M[block_ids[dep]][block_ids[k]] = 1

    # get initial nodes without dependencies
    frontier = deque(np.where(adj_M.sum(axis=0) == 0)[0].tolist())
    sorted_dag = []

    while frontier:
        elt = frontier.pop()
        sorted_dag.append(elt)
        dependants = np.where(adj_M[elt, :])[0]
        for dep in dependants:
            adj_M[elt, dep] = 0
            if not adj_M.sum(axis=0)[dep]:
                frontier.append(dep)

    if adj_M.sum() != 0:
        raise Exception("Cycle detected in analysis blocks dependencies, please review and try again!")

    sorted_blocks = []
    for idx in sorted_dag:
        sorted_blocks.append(block_objs[idx2block[idx]])

    return sorted_blocks


def _add_cls_kwarg(cls: Callable, kwargs: dict, key: str, value):
    """
    Adds arguments to the `kwargs` dictionary if the key-value pair is valid for the `cls` class signature.
    """
    if key in [p.name for p in inspect.signature(cls).parameters.values()]:
        kwargs[key] = value

    return kwargs



"""
Inductive conformal predictors.
"""
# Original author: Henrik Linusson (github.com/donlnz)
from collections import defaultdict
from functools import partial
from typing import Optional, Union
import numpy as np
from sklearn.base import BaseEstimator
from lightwood.analysis.nc.base import RegressorMixin, ClassifierMixin, TSMixin
from types import FunctionType


# -----------------------------------------------------------------------------
# Base inductive conformal predictor
# -----------------------------------------------------------------------------
class BaseIcp(BaseEstimator):
    """Base class for inductive conformal predictors.
    """

    def __init__(self, nc_function: FunctionType, condition: Union[bool, FunctionType] = None, cal_size: int = None):
        self.cal_x, self.cal_y = None, None
        self.nc_function = nc_function
        self.cal_size = cal_size  # if specified, defines size of calibration set

        # Check if condition-parameter is the default function (i.e.,
        # lambda x: 0). This is so we can safely clone the object without
        # the clone accidentally having self.conditional = True.
        def default_condition(x):
            return 0
        is_default = (callable(condition) and
                      (condition.__code__.co_code ==
                       default_condition.__code__.co_code))

        if is_default:
            self.condition = condition
            self.conditional = False
        elif callable(condition):
            self.condition = condition
            self.conditional = True
        else:
            self.condition = lambda x: 0
            self.conditional = False

    def fit(self, x: np.array, y: np.array) -> None:
        """Fit underlying nonconformity scorer.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for fitting the nonconformity scorer.

        y : numpy array of shape [n_samples]
            Outputs of examples for fitting the nonconformity scorer.

        Returns
        -------
        None
        """
        # TODO: incremental?
        self.nc_function.fit(x, y)

    def calibrate(self, x, y, increment=False):
        """Calibrate conformal predictor based on underlying nonconformity
        scorer.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for calibrating the conformal predictor.

        y : numpy array of shape [n_samples, n_features]
            Outputs of examples for calibrating the conformal predictor.

        increment : boolean
            If ``True``, performs an incremental recalibration of the conformal
            predictor. The supplied ``x`` and ``y`` are added to the set of
            previously existing calibration examples, and the conformal
            predictor is then calibrated on both the old and new calibration
            examples.

        Returns
        -------
        None
        """
        self._calibrate_hook(x, y, increment)
        self._update_calibration_set(x, y, increment)

        if self.conditional:
            category_map = np.array([self.condition((x[i, :], y[i]))
                                     for i in range(y.size)])
            self.categories = np.unique(category_map)
            self.cal_scores = defaultdict(partial(np.ndarray, 0))

            for cond in self.categories:
                idx = category_map == cond
                cal_scores = self.nc_function.score(self.cal_x[idx, :],
                                                    self.cal_y[idx])
                self.cal_scores[cond] = np.sort(cal_scores)[::-1]
        else:
            self.categories = np.array([0])
            cal_scores = self.nc_function.score(self.cal_x, self.cal_y)
            self.cal_scores = {0: np.sort(cal_scores)[::-1]}

        if self.cal_size:
            self.cal_scores = self._reduce_scores()

    def _reduce_scores(self):
        return {k: cs[::int(len(cs) / self.cal_size) + 1] for k, cs in self.cal_scores.items()}

    def _update_calibration_set(self, x: np.array, y: np.array, increment: bool) -> None:
        if increment and self.cal_x is not None and self.cal_y is not None:
            self.cal_x = np.vstack([self.cal_x, x])
            self.cal_y = np.hstack([self.cal_y, y])
        else:
            self.cal_x, self.cal_y = x, y

    def _calibrate_hook(self, x: np.array, y: np.array, increment: bool) -> None:
        pass


# -----------------------------------------------------------------------------
# Inductive conformal classifier
# -----------------------------------------------------------------------------
class IcpClassifier(BaseIcp, ClassifierMixin):
    """Inductive conformal classifier.

    Parameters
    ----------
    nc_function : BaseScorer
        Nonconformity scorer object used to calculate nonconformity of
        calibration examples and test patterns. Should implement ``fit(x, y)``
        and ``calc_nc(x, y)``.

    smoothing : boolean
        Decides whether to use stochastic smoothing of p-values.

    Attributes
    ----------
    cal_x : numpy array of shape [n_cal_examples, n_features]
        Inputs of calibration set.

    cal_y : numpy array of shape [n_cal_examples]
        Outputs of calibration set.

    nc_function : BaseScorer
        Nonconformity scorer object used to calculate nonconformity scores.

    classes : numpy array of shape [n_classes]
        List of class labels, with indices corresponding to output columns
         of IcpClassifier.predict()

    See also
    --------
    IcpRegressor

    References
    ----------
    .. [1] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
        intervals with regression neural networks. Neural Networks, 24(8),
        842-851.
    """

    def __init__(self, nc_function: FunctionType, condition: Union[bool, FunctionType] = None, cal_size: int = None,
                 smoothing: bool = True) -> None:
        super(IcpClassifier, self).__init__(nc_function, condition, cal_size)
        self.classes = None
        self.smoothing = smoothing

    def _calibrate_hook(self, x: np.array, y: np.array, increment: bool = False) -> None:
        self._update_classes(y, increment)

    def _update_classes(self, y: np.array, increment: bool) -> None:
        if self.classes is None or not increment:
            self.classes = np.unique(y)
        else:
            self.classes = np.unique(np.hstack([self.classes, y]))

    def predict(self, x: np.array, significance: Optional[float] = None) -> np.array:
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        significance : float or None
            Significance level (maximum allowed error rate) of predictions.
            Should be a float between 0 and 1. If ``None``, then the p-values
            are output rather than the predictions.

        Returns
        -------
        p : numpy array of shape [n_samples, n_classes]
            If significance is ``None``, then p contains the p-values for each
            sample-class pair; if significance is a float between 0 and 1, then
            p is a boolean array denoting which labels are included in the
            prediction sets.
        """
        # TODO: if x == self.last_x ...
        n_test_objects = x.shape[0]
        p = np.zeros((n_test_objects, self.classes.size))

        for i, c in enumerate(self.classes):
            test_class = np.zeros(x.shape[0], dtype=self.classes.dtype)
            test_class.fill(c)

            # TODO: maybe calculate p-values using cython or similar
            # TODO: interpolated p-values

            # TODO: nc_function.calc_nc should take X * {y1, y2, ... ,yn}
            test_nc_scores = self.nc_function.score(x, test_class)
            for j, nc in enumerate(test_nc_scores):
                cal_scores = self.cal_scores[self.condition((x[j, :], c))][::-1]
                n_cal = cal_scores.size
                n_eq = sum(np.where(cal_scores == nc, 1, 0))
                n_gt = sum(np.where(cal_scores > nc, 1, 0))

                if self.smoothing:
                    p[j, i] = (n_gt + n_eq * np.random.uniform(0, 1, 1)) / (n_cal + 1)
                else:
                    p[j, i] = (n_gt + n_eq) / (n_cal + 1)

        if significance is not None:
            return p > significance
        else:
            return p

    def predict_conf(self, x):
        """Predict the output values for a set of input patterns, using
        the confidence-and-credibility output scheme.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        Returns
        -------
        p : numpy array of shape [n_samples, 3]
            p contains three columns: the first column contains the most
            likely class for each test pattern; the second column contains
            the confidence in the predicted class label, and the third column
            contains the credibility of the prediction.
        """
        p = self.predict(x, significance=None)
        label = p.argmax(axis=1)
        credibility = p.max(axis=1)
        for i, idx in enumerate(label):
            p[i, idx] = -np.inf
        confidence = 1 - p.max(axis=1)

        return np.array([label, confidence, credibility]).T


# -----------------------------------------------------------------------------
# Inductive conformal regressor
# -----------------------------------------------------------------------------
class IcpRegressor(BaseIcp, RegressorMixin):
    """Inductive conformal regressor.

    Parameters
    ----------
    nc_function : BaseScorer
        Nonconformity scorer object used to calculate nonconformity of
        calibration examples and test patterns. Should implement ``fit(x, y)``,
        ``calc_nc(x, y)`` and ``predict(x, nc_scores, significance)``.

    Attributes
    ----------
    cal_x : numpy array of shape [n_cal_examples, n_features]
        Inputs of calibration set.

    cal_y : numpy array of shape [n_cal_examples]
        Outputs of calibration set.

    nc_function : BaseScorer
        Nonconformity scorer object used to calculate nonconformity scores.

    See also
    --------
    IcpClassifier

    References
    ----------
    .. [1] Papadopoulos, H., Proedrou, K., Vovk, V., & Gammerman, A. (2002).
        Inductive confidence machines for regression. In Machine Learning: ECML
        2002 (pp. 345-356). Springer Berlin Heidelberg.

    .. [2] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
        intervals with regression neural networks. Neural Networks, 24(8),
        842-851.
    """

    def __init__(self, nc_function: FunctionType, condition: bool = None, cal_size: int = None) -> None:
        super(IcpRegressor, self).__init__(nc_function, condition, cal_size)

    def predict(self, x: np.array, significance: bool = None) -> np.array:
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        significance : float
            Significance level (maximum allowed error rate) of predictions.
            Should be a float between 0 and 1. If ``None``, then intervals for
            all significance levels (0.01, 0.02, ..., 0.99) are output in a
            3d-matrix.

        Returns
        -------
        p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99}
            If significance is ``None``, then p contains the interval (minimum
            and maximum boundaries) for each test pattern, and each significance
            level (0.01, 0.02, ..., 0.99). If significance is a float between
            0 and 1, then p contains the prediction intervals (minimum and
            maximum	boundaries) for the set of test patterns at the chosen
            significance level.
        """
        # TODO: interpolated p-values

        n_significance = (99 if significance is None
                          else np.array(significance).size)

        if n_significance > 1:
            prediction = np.zeros((x.shape[0], 2, n_significance))
        else:
            prediction = np.zeros((x.shape[0], 2))

        condition_map = np.array([self.condition((x[i, :], None))
                                  for i in range(x.shape[0])])

        for condition in self.categories:
            idx = condition_map == condition
            if np.sum(idx) > 0:
                p = self.nc_function.predict(x[idx, :],
                                             self.cal_scores[condition],
                                             significance)
                if n_significance > 1:
                    prediction[idx, :, :] = p
                else:
                    prediction[idx, :] = p

        return prediction


# -----------------------------------------------------------------------------
# Inductive conformal forecasting
# -----------------------------------------------------------------------------
class IcpTSRegressor(BaseIcp, TSMixin):
    """Inductive conformal forecasting.

    Parameters
    ----------
    nc_function : BaseScorer
        Nonconformity scorer object used to calculate nonconformity of
        calibration examples and test patterns. Should implement ``fit(x, y)``,
        ``calc_nc(x, y)`` and ``predict(x, nc_scores, significance)``.

    Attributes
    ----------
    cal_x : numpy array of shape [n_cal_examples, n_features]
        Inputs of calibration set.

    cal_y : numpy array of shape [n_cal_examples, horizon_length]
        Outputs of calibration set.

    nc_function : BaseScorer
        Nonconformity scorer object used to calculate nonconformity scores.
    """

    def __init__(self, nc_function: FunctionType, horizon_length, condition: bool = None, cal_size: int = None) -> None:
        super(IcpTSRegressor, self).__init__(nc_function, condition, cal_size)
        self.horizon_length = horizon_length

    def calibrate(self, x, y, increment=False):
        """ 
        After calibration, handles incomplete target information by imputing the row-wise mean.
        """  # noqa
        super(IcpTSRegressor, self).calibrate(x, y, increment)
        for k, v in self.cal_scores.items():
            row_mean = np.nanmean(v, axis=1)
            idxs = np.where(np.isnan(v))
            v[idxs] = np.take(row_mean, idxs[0])
            self.cal_scores[k] = v

    def predict(self, x: np.array, significance: bool = None) -> np.array:
        """Predict the output values for a set of input patterns.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        significance : float
            Significance level (maximum allowed error rate) of predictions.
            Should be a float between 0 and 1. If ``None``, then intervals for
            all significance levels (0.01, 0.02, ..., 0.99) are output in a
            3d-matrix.

        Returns
        -------
        p : numpy array of shape [n_samples, horizon_length, 2] or [n_samples, horizon_length, 2, 99}
            If significance is ``None``, then p contains the interval (minimum
            and maximum boundaries) for each step of the test pattern, and each
            significance level (0.01, 0.02, ..., 0.99). If significance is a
            float between 0 and 1, then p contains the prediction intervals
            (minimum and maximum boundaries) for the all steps of the test
            patterns at the chosen significance level.
        """
        n_significance = (99 if significance is None else np.array(significance).size)

        if n_significance > 1:
            prediction = np.zeros((x.shape[0], self.horizon_length, 2, n_significance))
        else:
            prediction = np.zeros((x.shape[0], self.horizon_length, 2))

        condition_map = np.array([self.condition((x[i, :], None)) for i in range(x.shape[0])])

        for condition in self.categories:
            idx = condition_map == condition
            if np.sum(idx) > 0:
                p = self.nc_function.predict(x[idx, :],
                                             self.cal_scores[condition],
                                             significance)
                if n_significance > 1:
                    prediction[idx, :, :, :] = p
                else:
                    prediction[idx, :, :] = p

        return prediction



from typing import List, Optional

import torch
import pandas as pd

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.encoder.base import BaseEncoder
from lightwood.data.encoded_ds import EncodedDs
from lightwood.api.types import PredictionArguments


class Unit(BaseMixer):
    def __init__(self, stop_after: float, target_encoder: BaseEncoder):
        """
        The "Unit" mixer serves as a simple wrapper around a target encoder, essentially borrowing 
        the encoder's functionality for predictions. In other words, it simply arg-maxes the output of the encoder

        Used with encoders that already fine-tune on the targets (namely, pre-trained text ML models).
        
        Attributes:
            :param target_encoder: An instance of a Lightwood BaseEncoder. This encoder is used to decode predictions.
            :param stop_after (float): Time budget (in seconds) to train this mixer. 
        """  # noqa
        super().__init__(stop_after)
        self.target_encoder = target_encoder
        self.supports_proba = False
        self.stable = True

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        log.info("Unit mixer does not require training, it passes through predictions from its encoders.")

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        pass

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Makes predictions using the provided EncodedDs dataset.
        Mixer decodes predictions using the target encoder and returns them in a pandas DataFrame.

        :returns ydf (pd.DataFrame): a data frame containing the decoded predictions.
        """
        if args.predict_proba:
            # @TODO: depending on the target encoder, this might be enabled
            log.warning('This model does not output probability estimates')

        decoded_predictions: List[object] = []

        for X, _ in ds:
            decoded_prediction = self.target_encoder.decode(torch.unsqueeze(X, 0))
            decoded_predictions.extend(decoded_prediction)

        ydf = pd.DataFrame({"prediction": decoded_predictions})
        return ydf



from typing import Tuple, Dict
from datetime import datetime

import numpy as np
import pandas as pd


def get_ts_groups(df: pd.DataFrame, tss) -> list:
    group_combinations = ['__default']
    if tss.group_by:
        groups = [tuple([g]) if not isinstance(g, tuple) else g
                  for g in list(df.groupby(by=tss.group_by).groups.keys())]
        group_combinations.extend(groups)
    return group_combinations


def get_delta(df: pd.DataFrame, tss) -> Tuple[Dict, Dict, Dict]:
    """
    Infer the sampling interval of each time series, by picking the most popular time interval observed in the training data.

    :param df: Dataframe with time series data.
    :param tss: timeseries settings

    :return:
    Dictionary with group combination tuples as keys. Values are dictionaries with the inferred delta for each series.
    """  # noqa
    df = df.copy()  # TODO: necessary?
    original_col = f'__mdb_original_{tss.order_by}'
    order_col = original_col if original_col in df.columns else tss.order_by
    deltas = {"__default": df[order_col].astype(float).diff().value_counts().index[0]}
    freq, period = detect_freq_period(deltas["__default"], tss, len(df))
    periods = {"__default": [period]}
    freqs = {"__default": freq}

    if tss.group_by:
        grouped = df.groupby(by=tss.group_by)
        for group, subset in grouped:
            if subset.shape[0] > 1:
                deltas[group] = subset[order_col].diff().value_counts().index[0]
                freq, period = detect_freq_period(deltas[group], tss, len(subset))
                freqs[group] = freq
                periods[group] = [period] if period is not None else [1]
            else:
                deltas[group] = 1.0
                periods[group] = [1]
                freqs[group] = 'S'

    return deltas, periods, freqs


def get_inferred_timestamps(df: pd.DataFrame, col: str, deltas: dict, tss, stat_analysis,
                            time_format='') -> pd.DataFrame:
    horizon = tss.horizon

    last = np.vstack(df[f'order_{col}'].dropna().values)[:, -1]

    if tss.group_by:
        gby = [f'group_{g}' for g in tss.group_by]
        series_delta = df[gby].apply(lambda x: deltas.get(tuple(x.values.tolist()),
                                                          deltas['__default']), axis=1).values
        series_delta = series_delta.reshape(-1, 1)
    else:
        series_delta = np.full_like(df.values[:, 0:1], deltas['__default'])

    last = np.repeat(np.expand_dims(last, axis=1), horizon, axis=1)
    lins = np.linspace(0, horizon - 1, num=horizon)
    series_delta = np.repeat(series_delta, horizon, axis=1)
    timestamps = last + series_delta * lins

    if time_format:
        if time_format.lower() == 'infer':
            tformat = stat_analysis.ts_stats['order_format']
        else:
            tformat = time_format

        if tformat:
            def _strfts(elt):
                return datetime.utcfromtimestamp(elt).strftime(tformat)
            timestamps = np.vectorize(_strfts)(timestamps)

    # truncate to horizon
    timestamps = timestamps[:, :horizon]

    # preserves original input format if horizon == 1
    if tss.horizon == 1:
        timestamps = timestamps.squeeze()

    df[f'order_{col}'] = timestamps.tolist()
    return df[f'order_{col}']


def add_tn_num_conf_bounds(data: pd.DataFrame, tss_args):
    """
    Deprecated. Instead we now opt for the much better solution of having scores for each timestep (see all TS classes in analysis/nc)
    
    Add confidence (and bounds if applicable) to t+n predictions, for n>1
    TODO: active research question: how to guarantee 1-e coverage for t+n, n>1
    For now, (conservatively) increases width by the confidence times the log of the time step (and a scaling factor).
    """  # noqa
    for col in ['confidence', 'lower', 'upper']:
        data[col] = data[col].astype(object)

    for idx, row in data.iterrows():
        error_increase = [row['confidence'][0]] + \
                         [row['confidence'][0] * np.log(np.e + t / 2)  # offset by e so that y intercept is 1
                          for t in range(1, tss_args.horizon)]
        data['confidence'].iloc[idx] = [row['confidence'][0] for _ in range(tss_args.horizon)]

        preds = row['prediction']
        width = row['upper'][0] - row['lower'][0]
        data['lower'].iloc[idx] = [pred - (width / 2) * modifier for pred, modifier in zip(preds, error_increase)]
        data['upper'].iloc[idx] = [pred + (width / 2) * modifier for pred, modifier in zip(preds, error_increase)]

    return data


def add_tn_cat_conf_bounds(data: pd.DataFrame, tss_args):
    data['confidence'] = data['confidence'].astype(object)
    for idx, row in data.iterrows():
        data['confidence'].iloc[idx] = [row['confidence'] for _ in range(tss_args.horizon)]
    return data


class Differencer:
    def __init__(self):
        self.original_train_series = None
        self.diffed_train_series = None
        self.first_train_value = None
        self.last_train_value = None

    def diff(self, series: np.array) -> pd.Series:
        series = self._flatten_series(series)
        s = pd.Series(series)
        return s.shift(-1) - s

    def fit(self, series: np.array) -> None:
        series = self._flatten_series(series)
        self.first_train_value = series[0]
        self.last_train_value = series[-1]
        self.original_train_series = series
        self.diffed_train_series = self.diff(series)

    def transform(self, series: np.array) -> pd.Series:
        series = self._flatten_series(series)
        return self.diff(series).shift(1).fillna(0)

    def inverse_transform(self, series: pd.Series, init=None) -> pd.Series:
        origin = init if init else self.last_train_value
        s = pd.Series(origin)
        s = s.append(series).dropna()
        return s.expanding().sum()

    @staticmethod
    def _flatten_series(series: np.ndarray) -> np.ndarray:
        if len(series.shape) > 2:
            raise Exception(f"Input data should be shaped (A,) or (A, 1), got {series.shape}")
        elif len(series.shape) == 2:
            series = series.flatten()
        return series


def detect_freq_period(deltas: pd.DataFrame, tss, n_points) -> tuple:
    """
    Helper method that, based on the most popular interval for a time series, determines its seasonal peridiocity (sp).
    This bit of information can be crucial for good modelling with methods like ARIMA.

    Supported time intervals are:
        * 'year'
        * 'semestral'
        * 'quarter'
        * 'bimonthly'
        * 'monthly'
        * 'weekly'
        * 'daily'
        * 'hourly'
        * 'minute'
        * 'second'

    Note: all computations assume that the first provided `order_by` column is the one that specifies the sp.

    :param deltas: output of `get_delta`, has the most popular interval for each time series.
    :param tss: timeseries settings.

    :return: for all time series 1) a dictionary with its sp and 2) a dictionary with the detected sampling frequency
    """  # noqa
    secs_to_interval = {
        'yearly': 60 * 60 * 24 * 365,
        'quarterly': 60 * 60 * 24 * 365 // 4,
        'bimonthly': 60 * 60 * 24 * 31 * 2,
        'monthly': 60 * 60 * 24 * 31,
        'weekly': 60 * 60 * 24 * 7,
        'daily': 60 * 60 * 24,
        'hourly': 60 * 60,
        'minute': 60,
        'second': 1,
        'millisecond': 0.001,
        'microsecond': 1e-6,
        'nanosecond': 1e-9,
        'constant': 0
    }
    freq_to_period = {interval: period for (interval, period) in tss.interval_periods}
    for tag, period in (('yearly', 1), ('quarterly', 4), ('bimonthly', 6), ('monthly', 12),
                        ('weekly', 52), ('daily', 7), ('hourly', 24), ('minute', 1), ('second', 1), ('constant', 0)):
        if tag not in freq_to_period.keys():
            if period <= n_points:
                freq_to_period[tag] = period
            else:
                freq_to_period[tag] = None

    diffs = [(tag, abs(deltas - secs)) for tag, secs in secs_to_interval.items()]
    freq, min_diff = sorted(diffs, key=lambda x: x[1])[0]
    multiplier = 1
    if secs_to_interval[freq]:
        multiplier += int(min_diff / secs_to_interval[freq])
    return freq_to_pandas(freq, multiplier=multiplier), freq_to_period.get(freq, 1)


def freq_to_pandas(freq, multiplier=1):
    mapping = {
        'constant': 'N',
        'nanosecond': 'N',
        'microsecond': 'us',
        'millisecond': 'ms',
        'second': 'S',
        'minute': 'T',
        'hourly': 'H',  # custom logic
        'daily': 'D',  # custom logic
        'weekly': 'W',  # anchor logic
        'monthly': 'M',  # custom logic
        'bimonthly': 'M',
        'quarterly': 'Q',  # anchor and custom logic
        'yearly': 'Y',  # anchor and custom logic
    }

    # TODO: implement custom dispatch for better precision, use row sample if available:
    #  pandas.pydata.org/docs/user_guide/timeseries.html
    items = [mapping[freq]]
    if multiplier > 1:
        items.insert(0, str(multiplier))
    return ''.join(items)


def filter_ts(df: pd.DataFrame, tss, n_rows=1):
    """
    This method triggers only for timeseries datasets.

    It returns a dataframe that filters out all but the first ``n_rows`` per group.
    """  # noqa
    if tss.is_timeseries:
        gby = tss.group_by
        if gby is None:
            df = df.iloc[[0]]
        else:
            ndf = pd.DataFrame(columns=df.columns)
            grouped = df.groupby(by=tss.group_by)
            for group, subdf in grouped:
                if group != '__default':
                    ndf = pd.concat([ndf, subdf.iloc[:n_rows]])
            df = ndf
    return df



import requests
from lightwood.helpers.log import log
import os


def read_from_path_or_url(path: str, load_from_path):
    if path.startswith('http'):
        response = requests.get(path)
        with open(path.split('/')[-1], 'wb') as f:
            f.write(response.content)
        try:
            return load_from_path(path.split('/')[-1])
        except Exception as e:
            log.error(e)
        finally:
            os.remove(path.split('/')[-1])
    else:
        # Will automatically resample to 22.05kHz and convert to mono
        return load_from_path(path)




from lightwood.ensemble.base import BaseEnsemble
from lightwood.ensemble.identity import IdentityEnsemble
from lightwood.ensemble.embed import Embedder
from lightwood.ensemble.best_of import BestOf
from lightwood.ensemble.mean_ensemble import MeanEnsemble
from lightwood.ensemble.mode_ensemble import ModeEnsemble
from lightwood.ensemble.stacked_ensemble import StackedEnsemble
from lightwood.ensemble.ts_stacked_ensemble import TsStackedEnsemble
from lightwood.ensemble.weighted_mean_ensemble import WeightedMeanEnsemble

__all__ = ['BaseEnsemble', 'BestOf', 'MeanEnsemble', 'ModeEnsemble', 'WeightedMeanEnsemble', 'StackedEnsemble',
           'TsStackedEnsemble', 'Embedder', 'IdentityEnsemble']



# This encoder is optional since it's underlying dependency (librosa) needs system dependencies
try:
    from lightwood.encoder.audio.mfcc import MFCCEncoder
except Exception:
    MFCCEncoder = None

__all__ = ['MFCCEncoder']



import time
from copy import deepcopy
from typing import Dict, Optional, List

import numpy as np
import pandas as pd

import torch
from torch import nn
from torch.cuda.amp import GradScaler
from torch.utils.data import DataLoader

from type_infer.dtype import dtype
from lightwood.api.types import PredictionArguments
from lightwood.encoder.base import BaseEncoder
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs
from lightwood.mixer.neural import Neural
from lightwood.mixer.helpers.ar_net import ArNet
from lightwood.mixer.helpers.default_net import DefaultNet
from lightwood.api.types import TimeseriesSettings


class NeuralTs(Neural):
    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            timeseries_settings: TimeseriesSettings,
            target_encoder: BaseEncoder,
            net: str,
            fit_on_dev: bool,
            search_hyperparameters: bool,
            ts_analysis: Dict[str, Dict],
            n_epochs: Optional[int] = None,
            use_stl: Optional[bool] = False
    ):
        """
        Subclassed Neural mixer used for time series forecasting scenarios. 
        
        :param stop_after: How long the total fitting process should take
        :param target: Name of the target column
        :param dtype_dict: Data type dictionary
        :param timeseries_settings: TimeseriesSettings object for time-series tasks, refer to its documentation for available settings.
        :param target_encoder: Reference to the encoder used for the target
        :param net: The network type to use (`DeafultNet` or `ArNet`)
        :param fit_on_dev: If we should fit on the dev dataset
        :param search_hyperparameters: If the network should run a more through hyperparameter search (currently disabled)
        :param n_epochs: amount of epochs that the network will be trained for. Supersedes all other early stopping criteria if specified.
        """ # noqa
        super().__init__(
            stop_after,
            target,
            dtype_dict,
            target_encoder,
            net,
            fit_on_dev,
            search_hyperparameters,
            n_epochs,
        )
        self.timeseries_settings = timeseries_settings
        assert self.timeseries_settings.is_timeseries
        self.ts_analysis = ts_analysis
        self.net_class = DefaultNet if net == 'DefaultNet' else ArNet
        self.stable = True
        self.use_stl = use_stl

    def _select_criterion(self) -> torch.nn.Module:
        if self.dtype_dict[self.target] in (dtype.integer, dtype.float, dtype.num_tsarray, dtype.quantity):
            criterion = nn.L1Loss()
        else:
            criterion = super()._select_criterion()

        return criterion

    def _fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        :param train_data: The network is fit/trained on this
        :param dev_data: Data used for early stopping and hyperparameter determination
        """  # noqa
        self.started = time.time()
        original_train = deepcopy(train_data.data_frame)
        original_dev = deepcopy(dev_data.data_frame)

        # ConcatedEncodedDs
        self.batch_size = min(200, int(len(train_data) / 10))
        self.batch_size = max(40, self.batch_size)

        dev_dl = DataLoader(dev_data, batch_size=self.batch_size, shuffle=False)
        train_dl = DataLoader(train_data, batch_size=self.batch_size, shuffle=False)

        self.lr = 1e-4
        self.num_hidden = 1

        # Find learning rate
        # keep the weights
        self._init_net(train_data)
        self.lr, self.model = self._find_lr(train_data)

        # Keep on training
        optimizer = self._select_optimizer(self.model, lr=self.lr)
        criterion = self._select_criterion()
        scaler = GradScaler()

        # Only 0.8 of the remaining time budget is used to allow some time for the final tuning and partial fit
        self.model, epoch_to_best_model, _ = self._max_fit(
            train_dl, dev_dl, criterion, optimizer, scaler, (self.stop_after - (time.time() - self.started)) * 0.8,
            return_model_after=20000)

        self.epochs_to_best += epoch_to_best_model

        # restore dfs
        train_data.data_frame = original_train
        dev_data.data_frame = original_dev

        if self.fit_on_dev:
            self.partial_fit(dev_data, train_data)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        self._fit(train_data, dev_data)

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()
                 ) -> pd.DataFrame:
        original_df = deepcopy(ds.data_frame)

        self.model = self.model.eval()
        decoded_predictions = []
        all_probs: List[List[float]] = []
        rev_map = {}

        length = sum(ds.encoded_ds_lengths) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        pred_cols = [f'prediction_{i}' for i in range(self.timeseries_settings.horizon)]
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           dtype=object,
                           columns=pred_cols)

        with torch.no_grad():
            for idx, (X, Y) in enumerate(ds):
                X = X.to(self.model.device)
                Yh = self.model(X)
                Yh = torch.unsqueeze(Yh, 0) if len(Yh.shape) < 2 else Yh

                kwargs = {}
                for dep in self.target_encoder.dependencies:
                    kwargs['dependency_data'] = {dep: ds.data_frame.iloc[idx][[dep]].values}

                if args.predict_proba and self.supports_proba:
                    decoded_prediction, probs, rev_map = self.target_encoder.decode_probabilities(Yh, **kwargs)
                    all_probs.append(probs)
                else:
                    decoded_prediction = self.target_encoder.decode(Yh, **kwargs)

                decoded_predictions.extend(decoded_prediction)

        decoded_predictions = np.array(decoded_predictions)
        if len(decoded_predictions.shape) == 1:
            decoded_predictions = np.expand_dims(decoded_predictions, axis=1)
        ydf[pred_cols] = decoded_predictions

        ydf['prediction'] = ydf.values.tolist()

        if self.timeseries_settings.horizon == 1:
            ydf['prediction'] = [p[0] for p in ydf['prediction']]

        if args.predict_proba and self.supports_proba:
            raw_predictions = np.array(all_probs).squeeze(axis=1)

            for idx, label in enumerate(rev_map.values()):
                ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

        # TODO: make this part of the base mixer class? to avoid repetitive code
        #  and ensure other contribs don't accidentally modify the df
        ds.data_frame = original_df
        return ydf[['prediction']]



import warnings
from types import SimpleNamespace
from typing import Dict, Optional, Tuple

import numpy as np
import pandas as pd
from lightwood.analysis.base import BaseAnalysisBlock
from type_infer.dtype import dtype
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from lightwood.helpers.log import log
from sklearn.preprocessing import LabelEncoder

import shap


class ShapleyValues(BaseAnalysisBlock):
    """
    Analysis block that estimates column importance with SHAP (SHapley Additive exPlanations), a game theoretic approach
    to explain the ouput of any machine learning model. SHAP assigns each feature an importance value for a particular
    prediction.

    Reference:
        https://shap.readthedocs.io/en/stable/
        https://proceedings.neurips.cc/paper/2017/file/8a20a8621978632d76c43dfd28b67767-Paper.pdf
    """
    label_encoder: LabelEncoder

    def __init__(self, deps: Optional[Tuple] = ...):
        super().__init__(deps=deps)
        self.label_encoder = LabelEncoder()
        self.columns = []
        self.target = None

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        log.info('Preparing to compute feature importance values with SHAP')
        ns = SimpleNamespace(**kwargs)

        output_dtype = ns.dtype_dict[ns.target]
        train_data: EncodedDs = ns.train_data

        if output_dtype in (dtype.integer, dtype.float, dtype.quantity):
            pass
        elif output_dtype in (dtype.binary, dtype.categorical, dtype.tags):
            self.label_encoder.fit(train_data.data_frame[ns.target].values)
        else:
            log.warning(f'ShapleyValues analyzers not supported for type: {output_dtype}')
            return info

        self.target = ns.target
        self.columns = list(set(ns.dtype_dict.keys()) - {self.target})
        input_df = train_data.data_frame[self.columns]

        def model(x: np.ndarray) -> np.ndarray:
            assert(isinstance(x, np.ndarray))
            df = pd.DataFrame(data=x, columns=self.columns)
            ds = EncodedDs(encoders=train_data.encoders, data_frame=df, target=train_data.target)

            decoded_predictions = ns.predictor(ds=ds, args=PredictionArguments())
            if output_dtype in (dtype.integer, dtype.float, dtype.quantity):
                encoded_predictions = decoded_predictions['prediction'].values
            elif output_dtype in (dtype.binary, dtype.categorical, dtype.tags):
                encoded_predictions = self.label_encoder.transform(decoded_predictions['prediction'].values)

            return encoded_predictions

        info['shap_explainer'] = shap.KernelExplainer(model=model, data=input_df)

        return info

    def explain(self,
                row_insights: pd.DataFrame,
                global_insights: Dict[str, object],
                **kwargs
                ) -> Tuple[pd.DataFrame, Dict[str, object]]:
        log.info('Computing feature importance values with Kernel SHAP method')
        ns = SimpleNamespace(**kwargs)

        shap_explainer = ns.analysis.get('shap_explainer', None)
        if shap_explainer is None:
            return row_insights, global_insights

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            shap_values = shap_explainer.shap_values(ns.data[self.columns], silent=True)

        shap_values_df = pd.DataFrame(shap_values).rename(
            mapper=lambda i: f"shap_contribution_{self.columns[i]}", axis='columns')

        if kwargs.get('target_dtype', None) in (dtype.binary, dtype.categorical, dtype.tags):
            predictions = self.label_encoder.transform(row_insights['prediction'])
        else:
            predictions = row_insights['prediction']

        base_response = (predictions - shap_values_df.sum(axis='columns')).mean()

        row_insights = pd.concat([row_insights, shap_values_df], axis='columns')
        row_insights['shap_base_response'] = base_response
        row_insights['shap_final_response'] = predictions

        return row_insights, global_insights
